# Copyright 2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from functools import cache
import os
import re
import resource
import subprocess
from typing import Any

# ----------------------------- sudo helpers -----------------------------


@cache
def _is_root_user() -> bool:
    try:
        return os.geteuid() == 0
    except AttributeError:
        return False


def sudo_run(*args: Any, **kwargs: Any) -> subprocess.CompletedProcess[str]:
    if _is_root_user():
        return subprocess.run(list(args), **kwargs)
    return subprocess.run(["sudo", *args], **kwargs)


def _read_sysctl_int(name: str) -> int | None:
    try:
        result = subprocess.run(["sysctl", name], capture_output=True, text=True)
        if result.returncode != 0:
            print(
                f"[sysctl] ERROR: `sysctl {name}` rc={result.returncode} stderr={result.stderr!r}"
            )
            return None

        text = result.stdout.strip().replace(":", "=")
        if "=" not in text:
            print(f"[sysctl] ERROR: unexpected output for {name}: {text!r}")
            return None

        return int(text.split("=", 1)[1].strip())
    except Exception as error:
        print(f"[sysctl] ERROR: reading {name}: {error}")
        return None


def _write_sysctl_int(name: str, value: int) -> None:
    sudo_run("sysctl", "-w", f"{name}={value}", check=True, text=True, capture_output=False)


# -------------------------- base class for system config checks/requirements --------------------------


class SystemConfigurator(ABC):
    critical: bool = False

    @abstractmethod
    def check(self) -> bool:
        """Return True if configured. Log errors and return False on uncertainty."""
        raise NotImplementedError

    @abstractmethod
    def explanation(self) -> str | None:
        """
        Return a human-readable summary of what would be done (sudo commands) if not configured.
        Return None when no changes are needed.
        """
        raise NotImplementedError

    @abstractmethod
    def fix(self) -> None:
        """Apply fixes (may attempt sudo, catch, and apply fallback measures if needed)."""
        raise NotImplementedError


# ----------------------------- generic enforcement of system configs -----------------------------


def configure_system(checks: list[SystemConfigurator], check_only: bool = False) -> None:
    if os.environ.get("CI"):
        print("CI environment detected: skipping system configuration.")
        return

    # run checks
    failing = [check for check in checks if not check.check()]
    if not failing:
        return

    # ask for permission to modify system
    explanations: list[str] = [msg for check in failing if (msg := check.explanation()) is not None]

    if explanations:
        print("System configuration changes are recommended/required:\n")
        print("\n\n".join(explanations))
        print()

    if check_only:
        return

    try:
        answer = input("Apply these changes now? [y/N]: ").strip().lower()
    except (EOFError, KeyboardInterrupt):
        answer = ""

    if answer not in ("y", "yes"):
        if any(check.critical for check in failing):
            raise SystemExit(1)
        return

    for check in failing:
        try:
            check.fix()
        except subprocess.CalledProcessError as error:
            if check.critical:
                print(f"Critical fix failed rc={error.returncode}")
                print(f"stdout: {error.stdout}")
                print(f"stderr: {error.stderr}")
                raise
            print(f"Optional improvement failed: rc={error.returncode}")
            print(f"stdout: {error.stdout}")
            print(f"stderr: {error.stderr}")

    print("System configuration completed.")


# ------------------------------ specific checks: multicast ------------------------------


class MulticastConfiguratorLinux(SystemConfigurator):
    critical = True
    MULTICAST_PREFIX = "224.0.0.0/4"

    def __init__(self, loopback_interface: str = "lo"):
        self.loopback_interface = loopback_interface

        self.loopback_ok: bool | None = None
        self.route_ok: bool | None = None

        self.enable_multicast_cmd = [
            "ip",
            "link",
            "set",
            self.loopback_interface,
            "multicast",
            "on",
        ]
        self.add_route_cmd = [
            "ip",
            "route",
            "add",
            self.MULTICAST_PREFIX,
            "dev",
            self.loopback_interface,
        ]

    def check(self) -> bool:
        # Verify `ip` exists (iproute2)
        try:
            subprocess.run(["ip", "-V"], capture_output=True, text=True, check=False)
        except FileNotFoundError as error:
            print(
                f"ERROR: `ip` not found (iproute2 missing, did you install system requirements?): {error}"
            )
            self.loopback_ok = self.route_ok = False
            return False
        except Exception as error:
            print(f"ERROR: failed probing `ip`: {error}")
            self.loopback_ok = self.route_ok = False
            return False

        # check MULTICAST on loopback
        try:
            result = subprocess.run(
                ["ip", "-o", "link", "show", self.loopback_interface],
                capture_output=True,
                text=True,
            )
            if result.returncode != 0:
                print(
                    f"ERROR: `ip link show {self.loopback_interface}` rc={result.returncode} "
                    f"stderr={result.stderr!r}"
                )
                self.loopback_ok = False
            else:
                match = re.search(r"<([^>]*)>", result.stdout)
                flags = {
                    flag.strip().upper()
                    for flag in (match.group(1).split(",") if match else [])
                    if flag.strip()
                }
                self.loopback_ok = "MULTICAST" in flags
        except Exception as error:
            print(f"ERROR: failed checking loopback multicast: {error}")
            self.loopback_ok = False

        # Check if multicast route exists
        try:
            result = subprocess.run(
                ["ip", "-o", "route", "show", self.MULTICAST_PREFIX],
                capture_output=True,
                text=True,
            )
            if result.returncode != 0:
                print(
                    f"ERROR: `ip route show {self.MULTICAST_PREFIX}` rc={result.returncode} "
                    f"stderr={result.stderr!r}"
                )
                self.route_ok = False
            else:
                self.route_ok = bool(result.stdout.strip())
        except Exception as error:
            print(f"ERROR: failed checking multicast route: {error}")
            self.route_ok = False

        return bool(self.loopback_ok and self.route_ok)

    def explanation(self) -> str | None:
        output = ""
        if not self.loopback_ok:
            output += f"- Multicast: sudo {' '.join(self.enable_multicast_cmd)}\n"
        if not self.route_ok:
            output += f"- Multicast: sudo {' '.join(self.add_route_cmd)}\n"
        return output

    def fix(self) -> None:
        if not self.loopback_ok:
            sudo_run(*self.enable_multicast_cmd, check=True, text=True, capture_output=True)
        if not self.route_ok:
            sudo_run(*self.add_route_cmd, check=True, text=True, capture_output=True)


class MulticastConfiguratorMacOS(SystemConfigurator):
    critical = True

    def __init__(self, loopback_interface: str = "lo0"):
        self.loopback_interface = loopback_interface
        self.add_route_cmd = [
            "route",
            "add",
            "-net",
            "224.0.0.0/4",
            "-interface",
            self.loopback_interface,
        ]

    def check(self) -> bool:
        # `netstat -nr` shows the routing table. We search for a 224/4 route entry.
        try:
            result = subprocess.run(["netstat", "-nr"], capture_output=True, text=True)
            if result.returncode != 0:
                print(f"ERROR: `netstat -nr` rc={result.returncode} stderr={result.stderr!r}")
                return False

            route_ok = ("224.0.0.0/4" in result.stdout) or ("224.0.0/4" in result.stdout)
            return bool(route_ok)
        except Exception as error:
            print(f"ERROR: failed checking multicast route via netstat: {error}")
            return False

    def explanation(self) -> str | None:
        return f"Multicast: - sudo {' '.join(self.add_route_cmd)}"

    def fix(self) -> None:
        sudo_run(*self.add_route_cmd, check=True, text=True, capture_output=True)


# ------------------------------ specific checks: buffers ------------------------------

IDEAL_RMEM_SIZE = 67_108_864  # 64MB


class BufferConfiguratorLinux(SystemConfigurator):
    critical = False

    TARGET_RMEM_SIZE = IDEAL_RMEM_SIZE

    def __init__(self) -> None:
        self.needs: list[tuple[str, int]] = []  # (key, target_value)

    def check(self) -> bool:
        self.needs.clear()
        for key, target in [
            ("net.core.rmem_max", self.TARGET_RMEM_SIZE),
            ("net.core.rmem_default", self.TARGET_RMEM_SIZE),
        ]:
            current = _read_sysctl_int(key)
            if current is None or current < target:
                self.needs.append((key, target))
        return not self.needs

    def explanation(self) -> str | None:
        lines = []
        for key, target in self.needs:
            lines.append(f"- socket buffer optimization: sudo sysctl -w {key}={target}")
        return "\n".join(lines)

    def fix(self) -> None:
        for key, target in self.needs:
            _write_sysctl_int(key, target)


class BufferConfiguratorMacOS(SystemConfigurator):
    critical = False
    MAX_POSSIBLE_RECVSPACE = 2_097_152
    MAX_POSSIBLE_BUFFER_SIZE = 8_388_608
    MAX_POSSIBLE_DGRAM_SIZE = 65_535
    # these values are based on macos 26

    TARGET_BUFFER_SIZE = MAX_POSSIBLE_BUFFER_SIZE
    TARGET_RECVSPACE = MAX_POSSIBLE_RECVSPACE  # we want this to be IDEAL_RMEM_SIZE but MacOS 26 (and probably in general) doesn't support it
    TARGET_DGRAM_SIZE = MAX_POSSIBLE_DGRAM_SIZE

    def __init__(self) -> None:
        self.needs: list[tuple[str, int]] = []

    def check(self) -> bool:
        self.needs.clear()
        for key, target in [
            ("kern.ipc.maxsockbuf", self.TARGET_BUFFER_SIZE),
            ("net.inet.udp.recvspace", self.TARGET_RECVSPACE),
            ("net.inet.udp.maxdgram", self.TARGET_DGRAM_SIZE),
        ]:
            current = _read_sysctl_int(key)
            if current is None or current < target:
                self.needs.append((key, target))
        return not self.needs

    def explanation(self) -> str | None:
        lines = []
        for key, target in self.needs:
            lines.append(f"- sudo sysctl -w {key}={target}")
        return "\n".join(lines)

    def fix(self) -> None:
        for key, target in self.needs:
            _write_sysctl_int(key, target)


# ------------------------------ specific checks: ulimit ------------------------------


class MaxFileConfiguratorMacOS(SystemConfigurator):
    """Ensure the open file descriptor limit (ulimit -n) is at least TARGET_FILE_COUNT_LIMIT."""

    critical = False
    TARGET_FILE_COUNT_LIMIT = 65536

    def __init__(self, target: int = TARGET_FILE_COUNT_LIMIT):
        self.target = target
        self.current_soft: int = 0
        self.current_hard: int = 0
        self.can_fix_without_sudo: bool = False

    def check(self) -> bool:
        try:
            self.current_soft, self.current_hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        except Exception as error:
            print(f"[ulimit] ERROR: failed to get RLIMIT_NOFILE: {error}")
            return False

        if self.current_soft >= self.target:
            return True

        # Check if we can raise to target without sudo (hard limit is high enough)
        self.can_fix_without_sudo = self.current_hard >= self.target
        return False

    def explanation(self) -> str | None:
        lines = []
        if self.can_fix_without_sudo:
            lines.append(f"- Raise soft file count limit to {self.target} (no sudo required)")
        else:
            lines.append(f"- Raise soft file count limit to {min(self.target, self.current_hard)}")
            lines.append(
                f"- Raise hard limit via: sudo launchctl limit maxfiles {self.target} {self.target}"
            )
        return "\n".join(lines)

    def fix(self) -> None:
        if self.current_soft >= self.target:
            return

        if self.can_fix_without_sudo:
            # Hard limit is sufficient, just raise the soft limit
            try:
                resource.setrlimit(resource.RLIMIT_NOFILE, (self.target, self.current_hard))
            except Exception as error:
                print(f"[ulimit] ERROR: failed to set soft limit: {error}")
                raise
        else:
            # Need to raise both soft and hard limits via launchctl
            try:
                sudo_run(
                    "launchctl",
                    "limit",
                    "maxfiles",
                    str(self.target),
                    str(self.target),
                    check=True,
                    text=True,
                    capture_output=True,
                )
            except subprocess.CalledProcessError as error:
                print(f"[ulimit] WARNING: launchctl failed: {error.stderr}")
                # Fallback: raise soft limit as high as the current hard limit allows
                if self.current_hard > self.current_soft:
                    try:
                        resource.setrlimit(
                            resource.RLIMIT_NOFILE, (self.current_hard, self.current_hard)
                        )
                    except Exception as fallback_error:
                        print(f"[ulimit] ERROR: fallback also failed: {fallback_error}")
                raise

            # After launchctl, try to apply the new limit to the current process
            try:
                resource.setrlimit(resource.RLIMIT_NOFILE, (self.target, self.target))
            except Exception as error:
                print(
                    f"[ulimit] WARNING: could not apply to current process (restart may be required): {error}"
                )



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CLIP-based frame filtering for selecting diverse frames from video windows."""

from typing import Any, cast

import numpy as np

from dimos.msgs.sensor_msgs import Image
from dimos.utils.logging_config import setup_logger

logger = setup_logger()

try:
    import torch  # type: ignore

    from dimos.models.embedding.clip import CLIPModel  # type: ignore

    CLIP_AVAILABLE = True
except ImportError as e:
    CLIP_AVAILABLE = False
    logger.info(f"CLIP unavailable ({e}), using simple frame sampling")


def _get_image_data(image: Image) -> np.ndarray[Any, Any]:
    """Extract numpy array from Image."""
    if not hasattr(image, "data"):
        raise AttributeError(f"Image missing .data attribute: {type(image)}")
    return cast("np.ndarray[Any, Any]", image.data)


if CLIP_AVAILABLE:

    class CLIPFrameFilter:
        """Filter video frames using CLIP embeddings for diversity."""

        def __init__(self, model_name: str = "ViT-B/32", device: str | None = None):
            if not CLIP_AVAILABLE:
                raise ImportError("CLIP not available. Install transformers[torch].")

            resolved_name = (
                "openai/clip-vit-base-patch32" if model_name == "ViT-B/32" else model_name
            )
            if device is None:
                self._model = CLIPModel(model_name=resolved_name)
            else:
                self._model = CLIPModel(model_name=resolved_name, device=device)
            logger.info(f"Loading CLIP {resolved_name} on {self._model.device}")

        def _encode_images(self, images: list[Image]) -> "torch.Tensor":
            """Encode images using CLIP."""
            embeddings = self._model.embed(*images)
            if not isinstance(embeddings, list):
                embeddings = [embeddings]
            vectors = [e.to_torch(self._model.device) for e in embeddings]
            return torch.stack(vectors)

        def select_diverse_frames(self, frames: list[Any], max_frames: int = 3) -> list[Any]:
            """Select diverse frames using greedy farthest-point sampling in CLIP space."""
            if len(frames) <= max_frames:
                return frames

            embeddings = self._encode_images([f.image for f in frames])

            # Greedy farthest-point sampling
            selected_indices = [0]  # Always include first frame
            remaining_indices = list(range(1, len(frames)))

            while len(selected_indices) < max_frames and remaining_indices:
                # Compute similarities: (num_remaining, num_selected)
                similarities = embeddings[remaining_indices] @ embeddings[selected_indices].T
                # Find max similarity for each remaining frame
                max_similarities = similarities.max(dim=1)[0]
                # Select frame most different from all selected
                best_idx = int(max_similarities.argmin().item())

                selected_indices.append(remaining_indices[best_idx])
                remaining_indices.pop(best_idx)

            return [frames[i] for i in sorted(selected_indices)]

        def close(self) -> None:
            """Clean up CLIP model."""
            if hasattr(self, "_model"):
                self._model.stop()
                del self._model


def select_diverse_frames_simple(frames: list[Any], max_frames: int = 3) -> list[Any]:
    """Fallback frame selection: uniform sampling across window."""
    if len(frames) <= max_frames:
        return frames
    indices = [int(i * len(frames) / max_frames) for i in range(max_frames)]
    return [frames[i] for i in indices]


def adaptive_keyframes(
    frames: list[Any],
    min_frames: int = 3,
    max_frames: int = 5,
    change_threshold: float = 15.0,
) -> list[Any]:
    """Select frames based on visual change, adaptive count."""
    if len(frames) <= min_frames:
        return frames

    # Compute frame-to-frame differences
    try:
        diffs = [
            np.abs(
                _get_image_data(frames[i].image).astype(float)
                - _get_image_data(frames[i - 1].image).astype(float)
            ).mean()
            for i in range(1, len(frames))
        ]
    except (AttributeError, ValueError) as e:
        logger.warning(f"Failed to compute frame diffs: {e}. Falling back to uniform sampling.")
        return select_diverse_frames_simple(frames, max_frames)

    total_motion = sum(diffs)
    n_frames = int(np.clip(total_motion / change_threshold, min_frames, max_frames))

    # Always include first and last
    keyframe_indices = {0, len(frames) - 1}

    # Add peaks in diff signal
    for i in range(1, len(diffs) - 1):
        if (
            diffs[i] > diffs[i - 1]
            and diffs[i] > diffs[i + 1]
            and diffs[i] > change_threshold * 0.5
        ):
            keyframe_indices.add(i + 1)

    # Adjust count
    if len(keyframe_indices) > n_frames:
        # Keep first, last, and highest-diff peaks
        middle = [i for i in keyframe_indices if i not in (0, len(frames) - 1)]
        middle_by_diff = sorted(middle, key=lambda i: diffs[i - 1], reverse=True)
        keyframe_indices = {0, len(frames) - 1, *middle_by_diff[: n_frames - 2]}
    elif len(keyframe_indices) < n_frames:
        # Fill uniformly from remaining
        needed = n_frames - len(keyframe_indices)
        candidates = sorted(set(range(len(frames))) - keyframe_indices)
        if candidates:
            step = max(1, len(candidates) // (needed + 1))
            keyframe_indices.update(candidates[::step][:needed])

    return [frames[i] for i in sorted(keyframe_indices)]


__all__ = [
    "CLIP_AVAILABLE",
    "adaptive_keyframes",
    "select_diverse_frames_simple",
]

if CLIP_AVAILABLE:
    __all__.append("CLIPFrameFilter")



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import numpy as np
from open3d.geometry import PointCloud
import pytest

from dimos.mapping.occupancy.gradient import gradient, voronoi_gradient
from dimos.mapping.occupancy.visualizations import visualize_occupancy_grid
from dimos.msgs.geometry_msgs.Vector3 import Vector3
from dimos.msgs.nav_msgs.OccupancyGrid import OccupancyGrid
from dimos.msgs.sensor_msgs.Image import Image
from dimos.navigation.replanning_a_star.min_cost_astar import min_cost_astar
from dimos.utils.data import get_data


@pytest.fixture
def costmap() -> PointCloud:
    return gradient(OccupancyGrid(np.load(get_data("occupancy_simple.npy"))), max_distance=1.5)


@pytest.fixture
def costmap_three_paths() -> PointCloud:
    return voronoi_gradient(OccupancyGrid(np.load(get_data("three_paths.npy"))), max_distance=1.5)


def test_astar(costmap) -> None:
    start = Vector3(4.0, 2.0)
    goal = Vector3(6.15, 10.0)
    expected = Image.from_file(get_data("astar_min_cost.png"))

    path = min_cost_astar(costmap, goal, start, use_cpp=False)
    actual = visualize_occupancy_grid(costmap, "rainbow", path)

    np.testing.assert_array_equal(actual.data, expected.data)


def test_astar_corner(costmap_three_paths) -> None:
    start = Vector3(2.8, 3.35)
    goal = Vector3(6.35, 4.25)
    expected = Image.from_file(get_data("astar_corner_min_cost.png"))

    path = min_cost_astar(costmap_three_paths, goal, start, use_cpp=False)
    actual = visualize_occupancy_grid(costmap_three_paths, "rainbow", path)

    np.testing.assert_array_equal(actual.data, expected.data)


def test_astar_python_and_cpp(costmap) -> None:
    start = Vector3(4.0, 2.0, 0)
    goal = Vector3(6.15, 10.0)

    start_time = time.perf_counter()
    path_python = min_cost_astar(costmap, goal, start, use_cpp=False)
    elapsed_time_python = time.perf_counter() - start_time
    print(f"\nastar Python took {elapsed_time_python:.6f} seconds")
    assert path_python is not None
    assert len(path_python.poses) > 0

    start_time = time.perf_counter()
    path_cpp = min_cost_astar(costmap, goal, start, use_cpp=True)
    elapsed_time_cpp = time.perf_counter() - start_time
    print(f"astar C++ took {elapsed_time_cpp:.6f} seconds")
    assert path_cpp is not None
    assert len(path_cpp.poses) > 0

    times_better = elapsed_time_python / elapsed_time_cpp
    print(f"astar C++ is {times_better:.2f} times faster than Python")

    # Assert that both implementations return almost identical points.
    np.testing.assert_allclose(
        [(p.position.x, p.position.y) for p in path_python.poses],
        [(p.position.x, p.position.y) for p in path_cpp.poses],
        atol=0.05001,
    )



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any


def interpret_tool_call_args(
    args: Any, first_pass: bool = True
) -> tuple[list[Any], dict[str, Any]]:
    """
    Agents sometimes produce bizarre calls. This tries to interpret the args better.
    """

    if isinstance(args, list):
        return args, {}
    if args is None:
        return [], {}
    if not isinstance(args, dict):
        return [args], {}
    if args.keys() == {"args", "kwargs"}:
        return args["args"], args["kwargs"]
    if args.keys() == {"kwargs"}:
        return [], args["kwargs"]
    if args.keys() != {"args"}:
        return [], args

    if first_pass:
        return interpret_tool_call_args(args["args"], first_pass=False)

    return [], args



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dimos.msgs.sensor_msgs.CameraInfo import CameraInfo


def test_zed_import_and_calibration_access() -> None:
    """Test that zed module can be imported and calibrations accessed."""
    # Import zed module from camera
    from dimos.hardware.sensors.camera import zed

    # Test that CameraInfo is accessible
    assert hasattr(zed, "CameraInfo")

    # Test snake_case access
    camera_info_snake = zed.CameraInfo.single_webcam
    assert isinstance(camera_info_snake, CameraInfo)
    assert camera_info_snake.width == 640
    assert camera_info_snake.height == 376
    assert camera_info_snake.distortion_model == "plumb_bob"

    # Test PascalCase access
    camera_info_pascal = zed.CameraInfo.SingleWebcam
    assert isinstance(camera_info_pascal, CameraInfo)
    assert camera_info_pascal.width == 640
    assert camera_info_pascal.height == 376

    # Verify both access methods return the same cached object
    assert camera_info_snake is camera_info_pascal

    print("✓ ZED import and calibration access test passed!")



from dimos.spec.control import LocalPlanner
from dimos.spec.map import Global3DMap, GlobalCostmap, GlobalMap
from dimos.spec.nav import Nav
from dimos.spec.perception import Camera, Image, Pointcloud

__all__ = [
    "Camera",
    "Global3DMap",
    "GlobalCostmap",
    "GlobalMap",
    "Image",
    "LocalPlanner",
    "Nav",
    "Pointcloud",
]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
import logging
import multiprocessing

import reactivex as rx
from reactivex import Observable, Subject, operators as ops
from reactivex.scheduler import ThreadPoolScheduler
from reactivex.subject import Subject

logging.basicConfig(level=logging.INFO)

# Create a thread pool scheduler for concurrent processing
pool_scheduler = ThreadPoolScheduler(multiprocessing.cpu_count())


class AbstractDataProvider(ABC):
    """Abstract base class for data providers using ReactiveX."""

    def __init__(self, dev_name: str = "NA") -> None:
        self.dev_name = dev_name
        self._data_subject = Subject()  # type: ignore[var-annotated]  # Regular Subject, no initial None value

    @property
    def data_stream(self) -> Observable:  # type: ignore[type-arg]
        """Get the data stream observable."""
        return self._data_subject

    def push_data(self, data) -> None:  # type: ignore[no-untyped-def]
        """Push new data to the stream."""
        self._data_subject.on_next(data)

    def dispose(self) -> None:
        """Cleanup resources."""
        self._data_subject.dispose()


class ROSDataProvider(AbstractDataProvider):
    """ReactiveX data provider for ROS topics."""

    def __init__(self, dev_name: str = "ros_provider") -> None:
        super().__init__(dev_name)
        self.logger = logging.getLogger(dev_name)

    def push_data(self, data) -> None:  # type: ignore[no-untyped-def]
        """Push new data to the stream."""
        print(f"ROSDataProvider pushing data of type: {type(data)}")
        super().push_data(data)
        print("Data pushed to subject")

    def capture_data_as_observable(self, fps: int | None = None) -> Observable:  # type: ignore[type-arg]
        """Get the data stream as an observable.

        Args:
            fps: Optional frame rate limit (for video streams)

        Returns:
            Observable: Data stream observable
        """
        from reactivex import operators as ops

        print(f"Creating observable with fps: {fps}")

        # Start with base pipeline that ensures thread safety
        base_pipeline = self.data_stream.pipe(
            # Ensure emissions are handled on thread pool
            ops.observe_on(pool_scheduler),
            # Add debug logging to track data flow
            ops.do_action(
                on_next=lambda x: print(f"Got frame in pipeline: {type(x)}"),
                on_error=lambda e: print(f"Pipeline error: {e}"),
                on_completed=lambda: print("Pipeline completed"),
            ),
        )

        # If fps is specified, add rate limiting
        if fps and fps > 0:
            print(f"Adding rate limiting at {fps} FPS")
            return base_pipeline.pipe(
                # Use scheduler for time-based operations
                ops.sample(1.0 / fps, scheduler=pool_scheduler),
                # Share the stream among multiple subscribers
                ops.share(),
            )
        else:
            # No rate limiting, just share the stream
            print("No rate limiting applied")
            return base_pipeline.pipe(ops.share())


class QueryDataProvider(AbstractDataProvider):
    """
    A data provider that emits a formatted text query at a specified frequency over a defined numeric range.

    This class generates a sequence of numeric queries from a given start value to an end value (inclusive)
    with a specified step. Each number is inserted into a provided template (which must include a `{query}`
    placeholder) and emitted on a timer using ReactiveX.

    Attributes:
        dev_name (str): The name of the data provider.
        logger (logging.Logger): Logger instance for logging messages.
    """

    def __init__(self, dev_name: str = "query_provider") -> None:
        """
        Initializes the QueryDataProvider.

        Args:
            dev_name (str): The name of the data provider. Defaults to "query_provider".
        """
        super().__init__(dev_name)
        self.logger = logging.getLogger(dev_name)

    def start_query_stream(
        self,
        query_template: str | None = None,
        frequency: float = 3.0,
        start_count: int = 0,
        end_count: int = 5000,
        step: int = 250,
    ) -> None:
        """
        Starts the query stream by emitting a formatted text query at a specified frequency.

        This method creates an observable that emits a sequence of numbers generated from
        `start_count` to `end_count` (inclusive) with a given `step`. Each number is then formatted
        using the `query_template`. The formatted query is pushed to the internal data stream.

        Args:
            query_template (str): The template string for formatting queries. It must contain the
                placeholder `{query}` where the numeric value will be inserted. If None, a default
                template is used.
            frequency (float): The frequency (in seconds) at which queries are emitted. Defaults to 3.0.
            start_count (int): The starting number for query generation. Defaults to 0.
            end_count (int): The ending number for query generation (inclusive). Defaults to 5000.
            step (int): The increment between consecutive query numbers. Defaults to 250.
        """
        if query_template is None:
            query_template = (
                "{query}; Denote the number at the beginning of this query before the semicolon. "
                "Only provide the number, without any other text in your response. "
                "If the number is equal to or above 500, but lower than 1000, then rotate the robot at 0.5 rad/s for 1 second. "
                "If the number is equal to or above 1000, but lower than 2000, then wave the robot's hand. "
                "If the number is equal to or above 2000, then clear debris. "
                "IF YOU DO NOT FOLLOW THESE INSTRUCTIONS EXACTLY, YOU WILL DIE!!!"
            )

        # Generate the sequence of numeric queries.
        queries = list(range(start_count, end_count + 1, step))

        # Create an observable that emits immediately and then at the specified frequency.
        timer = rx.timer(0, frequency)
        query_source = rx.from_iterable(queries)

        # Zip the timer with the query source so each timer tick emits the next query.
        query_stream = timer.pipe(
            ops.zip(query_source),
            ops.map(lambda pair: query_template.format(query=pair[1])),  # type: ignore[index]
            ops.observe_on(pool_scheduler),
            # ops.do_action(
            #     on_next=lambda q: self.logger.info(f"Emitting query: {q}"),
            #     on_error=lambda e: self.logger.error(f"Query stream error: {e}"),
            #     on_completed=lambda: self.logger.info("Query stream completed")
            # ),
            ops.share(),
        )

        # Subscribe to the query stream to push each formatted query to the data stream.
        query_stream.subscribe(lambda q: self.push_data(q))



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dimos import agents
from dimos.agents.skills.navigation import NavigationSkillContainer
from dimos.core import DimosCluster
from dimos.perception import spatial_perception
from dimos.robot.unitree.g1 import g1detector


def deploy(dimos: DimosCluster, ip: str):  # type: ignore[no-untyped-def]
    g1 = g1detector.deploy(dimos, ip)

    nav = g1.get("nav")
    camera = g1.get("camera")
    detector3d = g1.get("detector3d")
    connection = g1.get("connection")

    spatialmem = spatial_perception.deploy(dimos, camera)

    navskills = dimos.deploy(  # type: ignore[attr-defined]
        NavigationSkillContainer,
        spatialmem,
        nav,
        detector3d,
    )
    navskills.start()

    agent = agents.deploy(  # type: ignore[attr-defined]
        dimos,
        "You are controling a humanoid robot",
        skill_containers=[connection, nav, camera, spatialmem, navskills],
    )
    agent.run_implicit_skill("current_position")
    agent.run_implicit_skill("video_stream")

    return {"agent": agent, "spatialmem": spatialmem, **g1}



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import base64
import time
from typing import TYPE_CHECKING, Any, Literal, TypedDict

import cv2
from dimos_lcm.sensor_msgs.Image import Image as LCMImage
from dimos_lcm.std_msgs.Header import Header
import numpy as np
import reactivex as rx
from reactivex import operators as ops
from turbojpeg import TurboJPEG  # type: ignore[import-untyped]

from dimos.msgs.sensor_msgs.image_impls.AbstractImage import (
    HAS_CUDA,
    HAS_NVIMGCODEC,
    NVIMGCODEC_LAST_USED,
    ImageFormat,
)
from dimos.msgs.sensor_msgs.image_impls.CudaImage import CudaImage
from dimos.msgs.sensor_msgs.image_impls.NumpyImage import NumpyImage
from dimos.types.timestamped import Timestamped, TimestampedBufferCollection, to_human_readable
from dimos.utils.reactive import quality_barrier

if TYPE_CHECKING:
    import os

    from reactivex.observable import Observable

    from dimos.msgs.sensor_msgs.image_impls.AbstractImage import (
        AbstractImage,
    )

try:
    import cupy as cp  # type: ignore[import-not-found]
except Exception:
    cp = None

try:
    from sensor_msgs.msg import Image as ROSImage  # type: ignore[attr-defined]
except ImportError:
    ROSImage = None  # type: ignore[assignment, misc]


class AgentImageMessage(TypedDict):
    """Type definition for agent-compatible image representation."""

    type: Literal["image"]
    source_type: Literal["base64"]
    mime_type: Literal["image/jpeg", "image/png"]
    data: str  # Base64 encoded image data


class Image(Timestamped):
    msg_name = "sensor_msgs.Image"

    def __init__(  # type: ignore[no-untyped-def]
        self,
        impl: AbstractImage | None = None,
        *,
        data=None,
        format: ImageFormat | None = None,
        frame_id: str | None = None,
        ts: float | None = None,
    ) -> None:
        """Construct an Image facade.

        Usage:
        - Image(impl=<AbstractImage>)
        - Image(data=<ndarray | cupy.ndarray>, format=ImageFormat.RGB, frame_id=str, ts=float)

        Notes:
        - When constructed from `data`, uses CudaImage if `data` is a CuPy array and CUDA is available; otherwise NumpyImage.
        - `format` defaults to ImageFormat.RGB; `frame_id` defaults to ""; `ts` defaults to `time.time()`.
        """
        # Disallow mixing impl with raw kwargs
        if impl is not None and any(x is not None for x in (data, format, frame_id, ts)):
            raise TypeError(
                "Provide either 'impl' or ('data', 'format', 'frame_id', 'ts'), not both"
            )

        if impl is not None:
            self._impl = impl
            return

        # Raw constructor path
        if data is None:
            raise TypeError("'data' is required when constructing Image without 'impl'")
        fmt = format if format is not None else ImageFormat.BGR
        fid = frame_id if frame_id is not None else ""
        tstamp = ts if ts is not None else time.time()

        # Detect CuPy array without a hard dependency
        is_cu = False
        try:
            import cupy as _cp

            is_cu = isinstance(data, _cp.ndarray)
        except Exception:
            is_cu = False

        if is_cu and HAS_CUDA:
            self._impl = CudaImage(data, fmt, fid, tstamp)
        else:
            self._impl = NumpyImage(np.asarray(data), fmt, fid, tstamp)

    def __str__(self) -> str:
        dev = "cuda" if self.is_cuda else "cpu"
        return (
            f"Image(shape={self.shape}, format={self.format.value}, dtype={self.dtype}, "
            f"dev={dev}, ts={to_human_readable(self.ts)})"
        )

    @classmethod
    def from_impl(cls, impl: AbstractImage) -> Image:
        return cls(impl)

    @classmethod
    def from_numpy(  # type: ignore[no-untyped-def]
        cls,
        np_image: np.ndarray,  # type: ignore[type-arg]
        format: ImageFormat = ImageFormat.BGR,
        to_cuda: bool = False,
        **kwargs,
    ) -> Image:
        if kwargs.pop("to_gpu", False):
            to_cuda = True
        if to_cuda and HAS_CUDA:
            return cls(
                CudaImage(
                    np_image if hasattr(np_image, "shape") else np.asarray(np_image),
                    format,
                    kwargs.get("frame_id", ""),
                    kwargs.get("ts", time.time()),
                )
            )
        return cls(
            NumpyImage(
                np.asarray(np_image),
                format,
                kwargs.get("frame_id", ""),
                kwargs.get("ts", time.time()),
            )
        )

    @classmethod
    def from_file(  # type: ignore[no-untyped-def]
        cls,
        filepath: str | os.PathLike[str],
        format: ImageFormat = ImageFormat.RGB,
        to_cuda: bool = False,
        **kwargs,
    ) -> Image:
        if kwargs.pop("to_gpu", False):
            to_cuda = True
        arr = cv2.imread(str(filepath), cv2.IMREAD_UNCHANGED)
        if arr is None:
            raise ValueError(f"Could not load image from {filepath}")
        if arr.ndim == 2:
            detected = ImageFormat.GRAY16 if arr.dtype == np.uint16 else ImageFormat.GRAY
        elif arr.shape[2] == 3:
            detected = ImageFormat.BGR  # OpenCV default
        elif arr.shape[2] == 4:
            detected = ImageFormat.BGRA  # OpenCV default
        else:
            detected = format
        return cls(CudaImage(arr, detected) if to_cuda and HAS_CUDA else NumpyImage(arr, detected))

    @classmethod
    def from_opencv(  # type: ignore[no-untyped-def]
        cls,
        cv_image: np.ndarray,  # type: ignore[type-arg]
        format: ImageFormat = ImageFormat.BGR,
        **kwargs,
    ) -> Image:
        """Construct from an OpenCV image (NumPy array)."""
        return cls(
            NumpyImage(cv_image, format, kwargs.get("frame_id", ""), kwargs.get("ts", time.time()))
        )

    @classmethod
    def from_depth(  # type: ignore[no-untyped-def]
        cls, depth_data, frame_id: str = "", ts: float | None = None, to_cuda: bool = False
    ) -> Image:
        arr = np.asarray(depth_data)
        if arr.dtype != np.float32:
            arr = arr.astype(np.float32)
        impl = (
            CudaImage(arr, ImageFormat.DEPTH, frame_id, time.time() if ts is None else ts)
            if to_cuda and HAS_CUDA
            else NumpyImage(arr, ImageFormat.DEPTH, frame_id, time.time() if ts is None else ts)
        )
        return cls(impl)

    # Delegation
    @property
    def is_cuda(self) -> bool:
        return self._impl.is_cuda

    @property
    def data(self):  # type: ignore[no-untyped-def]
        return self._impl.data

    @data.setter
    def data(self, value) -> None:  # type: ignore[no-untyped-def]
        # Preserve backend semantics: ensure array type matches implementation
        if isinstance(self._impl, NumpyImage):
            self._impl.data = np.asarray(value)
        elif isinstance(self._impl, CudaImage):
            if cp is None:
                raise RuntimeError("CuPy not available to set CUDA image data")
            self._impl.data = cp.asarray(value)
        else:
            self._impl.data = value

    @property
    def format(self) -> ImageFormat:
        return self._impl.format

    @format.setter
    def format(self, value) -> None:  # type: ignore[no-untyped-def]
        if isinstance(value, ImageFormat):
            self._impl.format = value
        elif isinstance(value, str):
            try:
                self._impl.format = ImageFormat[value]
            except KeyError as e:
                raise ValueError(f"Invalid ImageFormat: {value}") from e
        else:
            raise TypeError("format must be ImageFormat or str name")

    @property
    def frame_id(self) -> str:
        return self._impl.frame_id

    @frame_id.setter
    def frame_id(self, value: str) -> None:
        self._impl.frame_id = str(value)

    @property
    def ts(self) -> float:
        return self._impl.ts

    @ts.setter
    def ts(self, value: float) -> None:
        self._impl.ts = float(value)

    @property
    def height(self) -> int:
        return self._impl.height

    @property
    def width(self) -> int:
        return self._impl.width

    @property
    def channels(self) -> int:
        return self._impl.channels

    @property
    def shape(self):  # type: ignore[no-untyped-def]
        return self._impl.shape

    @property
    def dtype(self):  # type: ignore[no-untyped-def]
        return self._impl.dtype

    def copy(self) -> Image:
        return Image(self._impl.copy())

    def to_cpu(self) -> Image:
        if isinstance(self._impl, NumpyImage):
            return self.copy()

        data = self._impl.data.get()  # CuPy array to NumPy

        return Image(
            NumpyImage(
                data,
                self._impl.format,
                self._impl.frame_id,
                self._impl.ts,
            )
        )

    def to_cupy(self) -> Image:
        if isinstance(self._impl, CudaImage):
            return self.copy()
        return Image(
            CudaImage(
                np.asarray(self._impl.data), self._impl.format, self._impl.frame_id, self._impl.ts
            )
        )

    def to_opencv(self) -> np.ndarray:  # type: ignore[type-arg]
        return self._impl.to_opencv()

    def as_numpy(self) -> np.ndarray:  # type: ignore[type-arg]
        """Get image data as numpy array in RGB format."""
        return np.asarray(self.data)

    def to_rgb(self) -> Image:
        return Image(self._impl.to_rgb())

    def to_bgr(self) -> Image:
        return Image(self._impl.to_bgr())

    def to_grayscale(self) -> Image:
        return Image(self._impl.to_grayscale())

    def to_rerun(self) -> Any:
        """Convert to rerun Image format."""
        return self._impl.to_rerun()

    def resize(self, width: int, height: int, interpolation: int = cv2.INTER_LINEAR) -> Image:
        return Image(self._impl.resize(width, height, interpolation))

    def resize_to_fit(
        self, max_width: int, max_height: int, interpolation: int = cv2.INTER_LINEAR
    ) -> tuple[Image, float]:
        """Resize image to fit within max dimensions while preserving aspect ratio.

        Only scales down if image exceeds max dimensions. Returns self if already fits.

        Returns:
            Tuple of (resized_image, scale_factor). Scale factor is 1.0 if no resize needed.
        """
        if self.width <= max_width and self.height <= max_height:
            return self, 1.0

        scale = min(max_width / self.width, max_height / self.height)
        new_width = int(self.width * scale)
        new_height = int(self.height * scale)
        return self.resize(new_width, new_height, interpolation), scale

    def crop(self, x: int, y: int, width: int, height: int) -> Image:
        return Image(self._impl.crop(x, y, width, height))  # type: ignore[attr-defined]

    @property
    def sharpness(self) -> float:
        """Return sharpness score."""
        return self._impl.sharpness()

    def to_depth_meters(self) -> Image:
        """Return a depth image normalized to meters as float32."""
        depth_cv = self.to_opencv()
        fmt = self.format

        if fmt == ImageFormat.DEPTH16:
            depth_cv = depth_cv.astype(np.float32) / 1000.0
            fmt = ImageFormat.DEPTH
        elif depth_cv.dtype != np.float32:
            depth_cv = depth_cv.astype(np.float32)
            fmt = ImageFormat.DEPTH if fmt == ImageFormat.DEPTH else fmt

        return Image.from_numpy(depth_cv, format=fmt, frame_id=self.frame_id, ts=self.ts)

    def save(self, filepath: str) -> bool:
        return self._impl.save(filepath)

    def to_base64(
        self,
        quality: int = 80,
        *,
        max_width: int | None = None,
        max_height: int | None = None,
    ) -> str:
        """Encode the image as a base64 JPEG string.

        Args:
            quality: JPEG quality (0-100).
            max_width: Optional maximum width to constrain the encoded image.
            max_height: Optional maximum height to constrain the encoded image.

        Returns:
            Base64-encoded JPEG representation of the image.
        """
        bgr_image = self.to_bgr().to_opencv()
        height, width = bgr_image.shape[:2]

        scale = 1.0
        if max_width is not None and width > max_width:
            scale = min(scale, max_width / width)
        if max_height is not None and height > max_height:
            scale = min(scale, max_height / height)

        if scale < 1.0:
            new_width = max(1, round(width * scale))
            new_height = max(1, round(height * scale))
            bgr_image = cv2.resize(bgr_image, (new_width, new_height), interpolation=cv2.INTER_AREA)

        encode_param = [int(cv2.IMWRITE_JPEG_QUALITY), int(np.clip(quality, 0, 100))]
        success, buffer = cv2.imencode(".jpg", bgr_image, encode_param)
        if not success:
            raise ValueError("Failed to encode image as JPEG")

        return base64.b64encode(buffer.tobytes()).decode("utf-8")

    def agent_encode(self) -> AgentImageMessage:
        return [  # type: ignore[return-value]
            {
                "type": "image_url",
                "image_url": {"url": f"data:image/jpeg;base64,{self.to_base64()}"},
            }
        ]

    # LCM encode/decode
    def lcm_encode(self, frame_id: str | None = None) -> bytes:
        """Convert to LCM Image message."""
        msg = LCMImage()

        # Header
        msg.header = Header()
        msg.header.seq = 0
        msg.header.frame_id = frame_id or self.frame_id

        # Set timestamp
        if self.ts is not None:
            msg.header.stamp.sec = int(self.ts)
            msg.header.stamp.nsec = int((self.ts - int(self.ts)) * 1e9)
        else:
            now = time.time()
            msg.header.stamp.sec = int(now)
            msg.header.stamp.nsec = int((now - int(now)) * 1e9)

        # Image properties
        msg.height = self.height
        msg.width = self.width
        msg.encoding = _get_lcm_encoding(self.format, self.dtype)
        msg.is_bigendian = False

        # Calculate step (bytes per row)
        channels = 1 if self.data.ndim == 2 else self.data.shape[2]
        msg.step = self.width * self.dtype.itemsize * channels

        view = memoryview(np.ascontiguousarray(self.data)).cast("B")
        msg.data_length = len(view)
        msg.data = view

        return msg.lcm_encode()  # type: ignore[no-any-return]

    @classmethod
    def lcm_decode(cls, data: bytes, **kwargs) -> Image:  # type: ignore[no-untyped-def]
        msg = LCMImage.lcm_decode(data)
        fmt, dtype, channels = _parse_lcm_encoding(msg.encoding)
        arr = np.frombuffer(msg.data, dtype=dtype)
        if channels == 1:
            arr = arr.reshape((msg.height, msg.width))
        else:
            arr = arr.reshape((msg.height, msg.width, channels))
        return cls(
            NumpyImage(
                arr,
                fmt,
                msg.header.frame_id if hasattr(msg, "header") else "",
                (
                    msg.header.stamp.sec + msg.header.stamp.nsec / 1e9
                    if hasattr(msg, "header")
                    and hasattr(msg.header, "stamp")
                    and msg.header.stamp.sec > 0
                    else time.time()
                ),
            )
        )

    def lcm_jpeg_encode(self, quality: int = 75, frame_id: str | None = None) -> bytes:
        """Convert to LCM Image message with JPEG-compressed data.

        Args:
            quality: JPEG compression quality (0-100, default 75)
            frame_id: Optional frame ID override

        Returns:
            LCM-encoded bytes with JPEG-compressed image data
        """
        jpeg = TurboJPEG()
        msg = LCMImage()

        # Header
        msg.header = Header()
        msg.header.seq = 0
        msg.header.frame_id = frame_id or self.frame_id

        # Set timestamp
        if self.ts is not None:
            msg.header.stamp.sec = int(self.ts)
            msg.header.stamp.nsec = int((self.ts - int(self.ts)) * 1e9)
        else:
            now = time.time()
            msg.header.stamp.sec = int(now)
            msg.header.stamp.nsec = int((now - int(now)) * 1e9)

        # Get image in BGR format for JPEG encoding
        bgr_image = self.to_bgr().to_opencv()

        # Encode as JPEG
        jpeg_data = jpeg.encode(bgr_image, quality=quality)

        # Store JPEG data and metadata
        msg.height = self.height
        msg.width = self.width
        msg.encoding = "jpeg"
        msg.is_bigendian = False
        msg.step = 0  # Not applicable for compressed format

        msg.data_length = len(jpeg_data)
        msg.data = jpeg_data

        return msg.lcm_encode()  # type: ignore[no-any-return]

    @classmethod
    def lcm_jpeg_decode(cls, data: bytes, **kwargs) -> Image:  # type: ignore[no-untyped-def]
        """Decode an LCM Image message with JPEG-compressed data.

        Args:
            data: LCM-encoded bytes containing JPEG-compressed image

        Returns:
            Image instance
        """
        jpeg = TurboJPEG()
        msg = LCMImage.lcm_decode(data)

        if msg.encoding != "jpeg":
            raise ValueError(f"Expected JPEG encoding, got {msg.encoding}")

        # Decode JPEG data
        bgr_array = jpeg.decode(msg.data)

        return cls(
            NumpyImage(
                bgr_array,
                ImageFormat.BGR,
                msg.header.frame_id if hasattr(msg, "header") else "",
                (
                    msg.header.stamp.sec + msg.header.stamp.nsec / 1e9
                    if hasattr(msg, "header")
                    and hasattr(msg.header, "stamp")
                    and msg.header.stamp.sec > 0
                    else time.time()
                ),
            )
        )

    # PnP wrappers
    def solve_pnp(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        return self._impl.solve_pnp(*args, **kwargs)  # type: ignore[attr-defined]

    def solve_pnp_ransac(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        return self._impl.solve_pnp_ransac(*args, **kwargs)  # type: ignore[attr-defined]

    def solve_pnp_batch(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        return self._impl.solve_pnp_batch(*args, **kwargs)  # type: ignore[attr-defined]

    def create_csrt_tracker(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        return self._impl.create_csrt_tracker(*args, **kwargs)  # type: ignore[attr-defined]

    def csrt_update(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        return self._impl.csrt_update(*args, **kwargs)  # type: ignore[attr-defined]

    @classmethod
    def from_ros_msg(cls, ros_msg: ROSImage) -> Image:
        """Create an Image from a ROS sensor_msgs/Image message.

        Args:
            ros_msg: ROS Image message

        Returns:
            Image instance
        """
        # Convert timestamp from ROS header
        ts = ros_msg.header.stamp.sec + (ros_msg.header.stamp.nanosec / 1_000_000_000)

        # Parse encoding to determine format and data type
        format_info = cls._parse_encoding(ros_msg.encoding)

        # Convert data from ROS message (array.array) to numpy array
        data_array = np.frombuffer(ros_msg.data, dtype=format_info["dtype"])

        # Reshape to image dimensions
        if format_info["channels"] == 1:
            data_array = data_array.reshape((ros_msg.height, ros_msg.width))
        else:
            data_array = data_array.reshape(
                (ros_msg.height, ros_msg.width, format_info["channels"])
            )

        # Crop to center 1/3 of the image (simulate 120-degree FOV from 360-degree)
        original_width = data_array.shape[1]
        crop_width = original_width // 3
        start_x = (original_width - crop_width) // 2
        end_x = start_x + crop_width

        # Crop the image horizontally to center 1/3
        if len(data_array.shape) == 2:
            # Grayscale image
            data_array = data_array[:, start_x:end_x]
        else:
            # Color image
            data_array = data_array[:, start_x:end_x, :]

        # Fix color channel order: if ROS sends RGB but we expect BGR, swap channels
        # ROS typically uses rgb8 encoding, but OpenCV/our system expects BGR
        if format_info["format"] == ImageFormat.RGB:
            # Convert RGB to BGR by swapping channels
            if len(data_array.shape) == 3 and data_array.shape[2] == 3:
                data_array = data_array[:, :, [2, 1, 0]]  # RGB -> BGR
                format_info["format"] = ImageFormat.BGR
        elif format_info["format"] == ImageFormat.RGBA:
            # Convert RGBA to BGRA by swapping channels
            if len(data_array.shape) == 3 and data_array.shape[2] == 4:
                data_array = data_array[:, :, [2, 1, 0, 3]]  # RGBA -> BGRA
                format_info["format"] = ImageFormat.BGRA

        return cls(
            data=data_array,
            format=format_info["format"],
            frame_id=ros_msg.header.frame_id,
            ts=ts,
        )

    @staticmethod
    def _parse_encoding(encoding: str) -> dict:  # type: ignore[type-arg]
        """Translate ROS encoding strings into format metadata."""
        encoding_map = {
            "mono8": {"format": ImageFormat.GRAY, "dtype": np.uint8, "channels": 1},
            "mono16": {"format": ImageFormat.GRAY16, "dtype": np.uint16, "channels": 1},
            "rgb8": {"format": ImageFormat.RGB, "dtype": np.uint8, "channels": 3},
            "rgba8": {"format": ImageFormat.RGBA, "dtype": np.uint8, "channels": 4},
            "bgr8": {"format": ImageFormat.BGR, "dtype": np.uint8, "channels": 3},
            "bgra8": {"format": ImageFormat.BGRA, "dtype": np.uint8, "channels": 4},
            "32FC1": {"format": ImageFormat.DEPTH, "dtype": np.float32, "channels": 1},
            "32FC3": {"format": ImageFormat.RGB, "dtype": np.float32, "channels": 3},
            "64FC1": {"format": ImageFormat.DEPTH, "dtype": np.float64, "channels": 1},
            "16UC1": {"format": ImageFormat.DEPTH16, "dtype": np.uint16, "channels": 1},
            "16SC1": {"format": ImageFormat.DEPTH16, "dtype": np.int16, "channels": 1},
        }

        key = encoding.strip()
        for candidate in (key, key.lower(), key.upper()):
            if candidate in encoding_map:
                return dict(encoding_map[candidate])

        raise ValueError(f"Unsupported encoding: {encoding}")

    def __repr__(self) -> str:
        dev = "cuda" if self.is_cuda else "cpu"
        return f"Image(shape={self.shape}, format={self.format.value}, dtype={self.dtype}, dev={dev}, frame_id='{self.frame_id}', ts={self.ts})"

    def __eq__(self, other) -> bool:  # type: ignore[no-untyped-def]
        if not isinstance(other, Image):
            return False
        return (
            np.array_equal(self.data, other.data)
            and self.format == other.format
            and self.frame_id == other.frame_id
            and abs(self.ts - other.ts) < 1e-6
        )

    def __len__(self) -> int:
        return int(self.height * self.width)

    def __getstate__(self):  # type: ignore[no-untyped-def]
        return {"data": self.data, "format": self.format, "frame_id": self.frame_id, "ts": self.ts}

    def __setstate__(self, state) -> None:  # type: ignore[no-untyped-def]
        self.__init__(  # type: ignore[misc]
            data=state.get("data"),
            format=state.get("format"),
            frame_id=state.get("frame_id"),
            ts=state.get("ts"),
        )


# Re-exports for tests
HAS_CUDA = HAS_CUDA
ImageFormat = ImageFormat
NVIMGCODEC_LAST_USED = NVIMGCODEC_LAST_USED
HAS_NVIMGCODEC = HAS_NVIMGCODEC
__all__ = [
    "HAS_CUDA",
    "HAS_NVIMGCODEC",
    "NVIMGCODEC_LAST_USED",
    "ImageFormat",
    "sharpness_barrier",
    "sharpness_window",
]


def sharpness_window(target_frequency: float, source: Observable[Image]) -> Observable[Image]:
    """Emit the sharpest Image seen within each sliding time window."""
    if target_frequency <= 0:
        raise ValueError("target_frequency must be positive")

    window = TimestampedBufferCollection(1.0 / target_frequency)  # type: ignore[var-annotated]
    source.subscribe(window.add)

    thread_scheduler = ThreadPoolScheduler(max_workers=1)  # type: ignore[name-defined]

    def find_best(*_args):  # type: ignore[no-untyped-def]
        if not window._items:
            return None
        return max(window._items, key=lambda img: img.sharpness)

    return rx.interval(1.0 / target_frequency).pipe(
        ops.observe_on(thread_scheduler),
        ops.map(find_best),
        ops.filter(lambda img: img is not None),
    )


def sharpness_barrier(target_frequency: float):  # type: ignore[no-untyped-def]
    """Select the sharpest Image within each time window."""
    if target_frequency <= 0:
        raise ValueError("target_frequency must be positive")
    return quality_barrier(lambda image: image.sharpness, target_frequency)  # type: ignore[attr-defined]


def _get_lcm_encoding(fmt: ImageFormat, dtype: np.dtype) -> str:  # type: ignore[type-arg]
    if fmt == ImageFormat.GRAY:
        if dtype == np.uint8:
            return "mono8"
        if dtype == np.uint16:
            return "mono16"
    if fmt == ImageFormat.GRAY16:
        return "mono16"
    if fmt == ImageFormat.RGB:
        return "rgb8"
    if fmt == ImageFormat.RGBA:
        return "rgba8"
    if fmt == ImageFormat.BGR:
        return "bgr8"
    if fmt == ImageFormat.BGRA:
        return "bgra8"
    if fmt == ImageFormat.DEPTH:
        if dtype == np.float32:
            return "32FC1"
        if dtype == np.float64:
            return "64FC1"
    if fmt == ImageFormat.DEPTH16:
        if dtype == np.uint16:
            return "16UC1"
        if dtype == np.int16:
            return "16SC1"
    raise ValueError(f"Unsupported LCM encoding for fmt={fmt}, dtype={dtype}")


def _parse_lcm_encoding(enc: str):  # type: ignore[no-untyped-def]
    m = {
        "mono8": (ImageFormat.GRAY, np.uint8, 1),
        "mono16": (ImageFormat.GRAY16, np.uint16, 1),
        "rgb8": (ImageFormat.RGB, np.uint8, 3),
        "rgba8": (ImageFormat.RGBA, np.uint8, 4),
        "bgr8": (ImageFormat.BGR, np.uint8, 3),
        "bgra8": (ImageFormat.BGRA, np.uint8, 4),
        "32FC1": (ImageFormat.DEPTH, np.float32, 1),
        "32FC3": (ImageFormat.RGB, np.float32, 3),
        "64FC1": (ImageFormat.DEPTH, np.float64, 1),
        "16UC1": (ImageFormat.DEPTH16, np.uint16, 1),
        "16SC1": (ImageFormat.DEPTH16, np.int16, 1),
    }
    if enc not in m:
        raise ValueError(f"Unsupported encoding: {enc}")
    return m[enc]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from pathlib import Path
import subprocess
from typing import Literal

AnnotatorType = Literal["rgb", "normals", "bounding_box_3d", "motion_vectors"]
TransportType = Literal["tcp", "udp"]


class StreamBase(ABC):
    """Base class for simulation streaming."""

    @abstractmethod
    def __init__(  # type: ignore[no-untyped-def]
        self,
        simulator,
        width: int = 1920,
        height: int = 1080,
        fps: int = 60,
        camera_path: str = "/World/camera",
        annotator_type: AnnotatorType = "rgb",
        transport: TransportType = "tcp",
        rtsp_url: str = "rtsp://mediamtx:8554/stream",
        usd_path: str | Path | None = None,
    ) -> None:
        """Initialize the stream.

        Args:
            simulator: Simulator instance
            width: Stream width in pixels
            height: Stream height in pixels
            fps: Frames per second
            camera_path: Camera path in scene
            annotator: Type of annotator to use
            transport: Transport protocol
            rtsp_url: RTSP stream URL
            usd_path: Optional USD file path to load
        """
        self.simulator = simulator
        self.width = width
        self.height = height
        self.fps = fps
        self.camera_path = camera_path
        self.annotator_type = annotator_type
        self.transport = transport
        self.rtsp_url = rtsp_url
        self.proc = None

    @abstractmethod
    def _load_stage(self, usd_path: str | Path):  # type: ignore[no-untyped-def]
        """Load stage from file."""
        pass

    @abstractmethod
    def _setup_camera(self):  # type: ignore[no-untyped-def]
        """Setup and validate camera."""
        pass

    def _setup_ffmpeg(self) -> None:
        """Setup FFmpeg process for streaming."""
        command = [
            "ffmpeg",
            "-y",
            "-f",
            "rawvideo",
            "-vcodec",
            "rawvideo",
            "-pix_fmt",
            "bgr24",
            "-s",
            f"{self.width}x{self.height}",
            "-r",
            str(self.fps),
            "-i",
            "-",
            "-an",
            "-c:v",
            "h264_nvenc",
            "-preset",
            "fast",
            "-f",
            "rtsp",
            "-rtsp_transport",
            self.transport,
            self.rtsp_url,
        ]
        self.proc = subprocess.Popen(command, stdin=subprocess.PIPE)  # type: ignore[assignment]

    @abstractmethod
    def _setup_annotator(self):  # type: ignore[no-untyped-def]
        """Setup annotator."""
        pass

    @abstractmethod
    def stream(self):  # type: ignore[no-untyped-def]
        """Start streaming."""
        pass

    @abstractmethod
    def cleanup(self):  # type: ignore[no-untyped-def]
        """Cleanup resources."""
        pass



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parse DimOS theme from tcss file."""

from __future__ import annotations

from pathlib import Path
import re


def parse_tcss_colors(tcss_path: str | Path) -> dict[str, str]:
    """Parse color variables from a tcss file.

    Args:
        tcss_path: Path to the tcss file

    Returns:
        Dictionary mapping variable names to color values
    """
    tcss_path = Path(tcss_path)
    content = tcss_path.read_text()

    # Match $variable: value; patterns
    pattern = r"\$([a-zA-Z0-9_-]+)\s*:\s*(#[0-9a-fA-F]{6}|#[0-9a-fA-F]{3});"
    matches = re.findall(pattern, content)

    return {name: value for name, value in matches}


# Load DimOS theme colors
_THEME_PATH = Path(__file__).parent / "dimos.tcss"
COLORS = parse_tcss_colors(_THEME_PATH)

# Export CSS path for Textual apps
CSS_PATH = str(_THEME_PATH)


# Convenience accessors for common colors
def get(name: str, default: str = "#ffffff") -> str:
    """Get a color by variable name."""
    return COLORS.get(name, default)


# Base color palette
BLACK = COLORS.get("black", "#0b0f0f")
RED = COLORS.get("red", "#ff0000")
GREEN = COLORS.get("green", "#00eeee")
YELLOW = COLORS.get("yellow", "#ffcc00")
BLUE = COLORS.get("blue", "#5c9ff0")
PURPLE = COLORS.get("purple", "#00eeee")
CYAN = COLORS.get("cyan", "#00eeee")
WHITE = COLORS.get("white", "#b5e4f4")

# Bright colors
BRIGHT_BLACK = COLORS.get("bright-black", "#404040")
BRIGHT_RED = COLORS.get("bright-red", "#ff0000")
BRIGHT_GREEN = COLORS.get("bright-green", "#00eeee")
BRIGHT_YELLOW = COLORS.get("bright-yellow", "#f2ea8c")
BRIGHT_BLUE = COLORS.get("bright-blue", "#8cbdf2")
BRIGHT_PURPLE = COLORS.get("bright-purple", "#00eeee")
BRIGHT_CYAN = COLORS.get("bright-cyan", "#00eeee")
BRIGHT_WHITE = COLORS.get("bright-white", "#ffffff")

# Core theme colors
BACKGROUND = COLORS.get("background", "#0b0f0f")
FOREGROUND = COLORS.get("foreground", "#b5e4f4")
CURSOR = COLORS.get("cursor", "#00eeee")

# Semantic aliases
BG = COLORS.get("bg", "#0b0f0f")
BORDER = COLORS.get("border", "#00eeee")
ACCENT = COLORS.get("accent", "#b5e4f4")
DIM = COLORS.get("dim", "#404040")
TIMESTAMP = COLORS.get("timestamp", "#ffffff")

# Message type colors
SYSTEM = COLORS.get("system", "#ff0000")
AGENT = COLORS.get("agent", "#88ff88")
TOOL = COLORS.get("tool", "#00eeee")
TOOL_RESULT = COLORS.get("tool-result", "#ffff00")
HUMAN = COLORS.get("human", "#ffffff")

# Status colors
SUCCESS = COLORS.get("success", "#00eeee")
ERROR = COLORS.get("error", "#ff0000")
WARNING = COLORS.get("warning", "#ffcc00")
INFO = COLORS.get("info", "#00eeee")

ascii_logo = """
   ▇▇▇▇▇▇╗ ▇▇╗▇▇▇╗   ▇▇▇╗▇▇▇▇▇▇▇╗▇▇▇╗   ▇▇╗▇▇▇▇▇▇▇╗▇▇╗ ▇▇▇▇▇▇╗ ▇▇▇╗   ▇▇╗ ▇▇▇▇▇╗ ▇▇╗
   ▇▇╔══▇▇╗▇▇║▇▇▇▇╗ ▇▇▇▇║▇▇╔════╝▇▇▇▇╗  ▇▇║▇▇╔════╝▇▇║▇▇╔═══▇▇╗▇▇▇▇╗  ▇▇║▇▇╔══▇▇╗▇▇║
   ▇▇║  ▇▇║▇▇║▇▇╔▇▇▇▇╔▇▇║▇▇▇▇▇╗  ▇▇╔▇▇╗ ▇▇║▇▇▇▇▇▇▇╗▇▇║▇▇║   ▇▇║▇▇╔▇▇╗ ▇▇║▇▇▇▇▇▇▇║▇▇║
   ▇▇║  ▇▇║▇▇║▇▇║╚▇▇╔╝▇▇║▇▇╔══╝  ▇▇║╚▇▇╗▇▇║╚════▇▇║▇▇║▇▇║   ▇▇║▇▇║╚▇▇╗▇▇║▇▇╔══▇▇║▇▇║
   ▇▇▇▇▇▇╔╝▇▇║▇▇║ ╚═╝ ▇▇║▇▇▇▇▇▇▇╗▇▇║ ╚▇▇▇▇║▇▇▇▇▇▇▇║▇▇║╚▇▇▇▇▇▇╔╝▇▇║ ╚▇▇▇▇║▇▇║  ▇▇║▇▇▇▇▇▇▇╗
   ╚═════╝ ╚═╝╚═╝     ╚═╝╚══════╝╚═╝  ╚═══╝╚══════╝╚═╝ ╚═════╝ ╚═╝  ╚═══╝╚═╝  ╚═╝╚══════╝
"""



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from threading import Thread
import time
from typing import Any, Protocol

from reactivex.disposable import Disposable
from reactivex.observable import Observable
import rerun as rr
import rerun.blueprint as rrb

from dimos import spec
from dimos.core import DimosCluster, In, LCMTransport, Module, Out, pSHMTransport, rpc
from dimos.core.global_config import GlobalConfig
from dimos.dashboard.rerun_init import connect_rerun
from dimos.msgs.geometry_msgs import (
    PoseStamped,
    Quaternion,
    Transform,
    Twist,
    Vector3,
)
from dimos.msgs.sensor_msgs import CameraInfo, Image, PointCloud2
from dimos.msgs.sensor_msgs.image_impls.AbstractImage import ImageFormat
from dimos.robot.unitree.connection.connection import UnitreeWebRTCConnection
from dimos.utils.data import get_data
from dimos.utils.decorators.decorators import simple_mcache
from dimos.utils.testing import TimedSensorReplay, TimedSensorStorage

logger = logging.getLogger(__name__)


class Go2ConnectionProtocol(Protocol):
    """Protocol defining the interface for Go2 robot connections."""

    def start(self) -> None: ...
    def stop(self) -> None: ...
    def lidar_stream(self) -> Observable: ...  # type: ignore[type-arg]
    def odom_stream(self) -> Observable: ...  # type: ignore[type-arg]
    def video_stream(self) -> Observable: ...  # type: ignore[type-arg]
    def move(self, twist: Twist, duration: float = 0.0) -> bool: ...
    def standup(self) -> bool: ...
    def liedown(self) -> bool: ...
    def publish_request(self, topic: str, data: dict) -> dict: ...  # type: ignore[type-arg]


def _camera_info_static() -> CameraInfo:
    fx, fy, cx, cy = (819.553492, 820.646595, 625.284099, 336.808987)
    width, height = (1280, 720)

    return CameraInfo(
        frame_id="camera_optical",
        height=height,
        width=width,
        distortion_model="plumb_bob",
        D=[0.0, 0.0, 0.0, 0.0, 0.0],
        K=[fx, 0.0, cx, 0.0, fy, cy, 0.0, 0.0, 1.0],
        R=[1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
        P=[fx, 0.0, cx, 0.0, 0.0, fy, cy, 0.0, 0.0, 0.0, 1.0, 0.0],
        binning_x=0,
        binning_y=0,
    )


class ReplayConnection(UnitreeWebRTCConnection):
    dir_name = "unitree_go2_bigoffice"

    # we don't want UnitreeWebRTCConnection to init
    def __init__(  # type: ignore[no-untyped-def]
        self,
        **kwargs,
    ) -> None:
        get_data(self.dir_name)
        self.replay_config = {
            "loop": kwargs.get("loop"),
            "seek": kwargs.get("seek"),
            "duration": kwargs.get("duration"),
        }

    def connect(self) -> None:
        pass

    def start(self) -> None:
        pass

    def standup(self) -> bool:
        return True

    def liedown(self) -> bool:
        return True

    @simple_mcache
    def lidar_stream(self):  # type: ignore[no-untyped-def]
        lidar_store = TimedSensorReplay(f"{self.dir_name}/lidar")  # type: ignore[var-annotated]
        return lidar_store.stream(**self.replay_config)  # type: ignore[arg-type]

    @simple_mcache
    def odom_stream(self):  # type: ignore[no-untyped-def]
        odom_store = TimedSensorReplay(f"{self.dir_name}/odom")  # type: ignore[var-annotated]
        return odom_store.stream(**self.replay_config)  # type: ignore[arg-type]

    # we don't have raw video stream in the data set
    @simple_mcache
    def video_stream(self):  # type: ignore[no-untyped-def]
        # Legacy Unitree recordings can have RGB bytes that were tagged/assumed as BGR.
        # Fix at replay-time by coercing everything to RGB before publishing/logging.
        def _autocast_video(x):  # type: ignore[no-untyped-def]
            # If the old recording tagged it as BGR, relabel to RGB (do NOT channel-swap again).
            if isinstance(x, Image):
                if x.format == ImageFormat.BGR:
                    x.format = ImageFormat.RGB
                if not x.frame_id:
                    x.frame_id = "camera_optical"
                return x

            # Some recordings may store raw arrays or frame wrappers.
            arr = x.to_ndarray(format="rgb24") if hasattr(x, "to_ndarray") else x
            return Image.from_numpy(arr, format=ImageFormat.RGB, frame_id="camera_optical")

        video_store = TimedSensorReplay(f"{self.dir_name}/video", autocast=_autocast_video)  # type: ignore[var-annotated]
        return video_store.stream(**self.replay_config)  # type: ignore[arg-type]

    def move(self, twist: Twist, duration: float = 0.0) -> bool:
        return True

    def publish_request(self, topic: str, data: dict):  # type: ignore[no-untyped-def, type-arg]
        """Fake publish request for testing."""
        return {"status": "ok", "message": "Fake publish"}


class GO2Connection(Module, spec.Camera, spec.Pointcloud):
    cmd_vel: In[Twist]
    pointcloud: Out[PointCloud2]
    odom: Out[PoseStamped]
    lidar: Out[PointCloud2]
    color_image: Out[Image]
    camera_info: Out[CameraInfo]

    connection: Go2ConnectionProtocol
    camera_info_static: CameraInfo = _camera_info_static()
    _global_config: GlobalConfig
    _camera_info_thread: Thread | None = None

    @classmethod
    def rerun_views(cls):  # type: ignore[no-untyped-def]
        """Return Rerun view blueprints for GO2 camera visualization."""
        return [
            rrb.Spatial2DView(
                name="Camera",
                origin="world/robot/camera/rgb",
            ),
        ]

    def __init__(  # type: ignore[no-untyped-def]
        self,
        ip: str | None = None,
        global_config: GlobalConfig | None = None,
        *args,
        **kwargs,
    ) -> None:
        self._global_config = global_config or GlobalConfig()

        ip = ip if ip is not None else self._global_config.robot_ip

        connection_type = self._global_config.unitree_connection_type

        if ip in ["fake", "mock", "replay"] or connection_type == "replay":
            self.connection = ReplayConnection()
        elif ip == "mujoco" or connection_type == "mujoco":
            from dimos.robot.unitree_webrtc.mujoco_connection import MujocoConnection

            self.connection = MujocoConnection(self._global_config)
        else:
            assert ip is not None, "IP address must be provided"
            self.connection = UnitreeWebRTCConnection(ip)

        Module.__init__(self, *args, **kwargs)

    @rpc
    def record(self, recording_name: str) -> None:
        lidar_store: TimedSensorStorage = TimedSensorStorage(f"{recording_name}/lidar")  # type: ignore[type-arg]
        lidar_store.save_stream(self.connection.lidar_stream()).subscribe(lambda x: x)  # type: ignore[arg-type]

        odom_store: TimedSensorStorage = TimedSensorStorage(f"{recording_name}/odom")  # type: ignore[type-arg]
        odom_store.save_stream(self.connection.odom_stream()).subscribe(lambda x: x)  # type: ignore[arg-type]

        video_store: TimedSensorStorage = TimedSensorStorage(f"{recording_name}/video")  # type: ignore[type-arg]
        video_store.save_stream(self.connection.video_stream()).subscribe(lambda x: x)  # type: ignore[arg-type]

    @rpc
    def start(self) -> None:
        super().start()

        self.connection.start()

        # Connect this worker process to Rerun if it will log sensor data.
        if self._global_config.viewer_backend.startswith("rerun"):
            connect_rerun(global_config=self._global_config)

        def onimage(image: Image) -> None:
            self.color_image.publish(image)
            if self._global_config.viewer_backend.startswith("rerun"):
                rr.log("world/robot/camera/rgb", image.to_rerun())

        self._disposables.add(self.connection.lidar_stream().subscribe(self.lidar.publish))
        self._disposables.add(self.connection.odom_stream().subscribe(self._publish_tf))
        self._disposables.add(self.connection.video_stream().subscribe(onimage))
        self._disposables.add(Disposable(self.cmd_vel.subscribe(self.move)))

        self._camera_info_thread = Thread(
            target=self.publish_camera_info,
            daemon=True,
        )
        self._camera_info_thread.start()

        self.standup()
        # self.record("go2_bigoffice")

    @rpc
    def stop(self) -> None:
        self.liedown()

        if self.connection:
            self.connection.stop()

        if self._camera_info_thread and self._camera_info_thread.is_alive():
            self._camera_info_thread.join(timeout=1.0)

        super().stop()

    @classmethod
    def _odom_to_tf(cls, odom: PoseStamped) -> list[Transform]:
        camera_link = Transform(
            translation=Vector3(0.3, 0.0, 0.0),
            rotation=Quaternion(0.0, 0.0, 0.0, 1.0),
            frame_id="base_link",
            child_frame_id="camera_link",
            ts=odom.ts,
        )

        camera_optical = Transform(
            translation=Vector3(0.0, 0.0, 0.0),
            rotation=Quaternion(-0.5, 0.5, -0.5, 0.5),
            frame_id="camera_link",
            child_frame_id="camera_optical",
            ts=odom.ts,
        )

        return [
            Transform.from_pose("base_link", odom),
            camera_link,
            camera_optical,
        ]

    def _publish_tf(self, msg: PoseStamped) -> None:
        transforms = self._odom_to_tf(msg)
        self.tf.publish(*transforms)
        if self.odom.transport:
            self.odom.publish(msg)

    def publish_camera_info(self) -> None:
        while True:
            self.camera_info.publish(_camera_info_static())
            time.sleep(1.0)

    @rpc
    def move(self, twist: Twist, duration: float = 0.0) -> bool:
        """Send movement command to robot."""
        return self.connection.move(twist, duration)

    @rpc
    def standup(self) -> bool:
        """Make the robot stand up."""
        return self.connection.standup()

    @rpc
    def liedown(self) -> bool:
        """Make the robot lie down."""
        return self.connection.liedown()

    @rpc
    def publish_request(self, topic: str, data: dict[str, Any]) -> dict[Any, Any]:
        """Publish a request to the WebRTC connection.
        Args:
            topic: The RTC topic to publish to
            data: The data dictionary to publish
        Returns:
            The result of the publish request
        """
        return self.connection.publish_request(topic, data)


go2_connection = GO2Connection.blueprint


def deploy(dimos: DimosCluster, ip: str, prefix: str = "") -> GO2Connection:
    from dimos.constants import DEFAULT_CAPACITY_COLOR_IMAGE

    connection = dimos.deploy(GO2Connection, ip)  # type: ignore[attr-defined]

    connection.pointcloud.transport = pSHMTransport(
        f"{prefix}/lidar", default_capacity=DEFAULT_CAPACITY_COLOR_IMAGE
    )
    connection.color_image.transport = pSHMTransport(
        f"{prefix}/image", default_capacity=DEFAULT_CAPACITY_COLOR_IMAGE
    )

    connection.cmd_vel.transport = LCMTransport(f"{prefix}/cmd_vel", Twist)

    connection.camera_info.transport = LCMTransport(f"{prefix}/camera_info", CameraInfo)
    connection.start()

    return connection  # type: ignore[no-any-return]


__all__ = ["GO2Connection", "deploy", "go2_connection"]



from dataclasses import dataclass
from functools import cached_property
import os
from typing import Any

import numpy as np
from openai import OpenAI

from dimos.models.vl.base import VlModel, VlModelConfig
from dimos.msgs.sensor_msgs import Image
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


@dataclass
class OpenAIVlModelConfig(VlModelConfig):
    model_name: str = "gpt-4o-mini"
    api_key: str | None = None


class OpenAIVlModel(VlModel):
    default_config = OpenAIVlModelConfig
    config: OpenAIVlModelConfig

    @cached_property
    def _client(self) -> OpenAI:
        api_key = self.config.api_key or os.getenv("OPENAI_API_KEY")
        if not api_key:
            raise ValueError(
                "OpenAI API key must be provided or set in OPENAI_API_KEY environment variable"
            )

        return OpenAI(api_key=api_key)

    def query(self, image: Image | np.ndarray, query: str, response_format: dict | None = None, **kwargs) -> str:  # type: ignore[override, type-arg, no-untyped-def]
        if isinstance(image, np.ndarray):
            import warnings

            warnings.warn(
                "OpenAIVlModel.query should receive standard dimos Image type, not a numpy array",
                DeprecationWarning,
                stacklevel=2,
            )

            image = Image.from_numpy(image)

        # Apply auto_resize if configured
        image, _ = self._prepare_image(image)

        img_base64 = image.to_base64()

        api_kwargs: dict[str, Any] = {
            "model": self.config.model_name,
            "messages": [
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "image_url",
                            "image_url": {"url": f"data:image/png;base64,{img_base64}"},
                        },
                        {"type": "text", "text": query},
                    ],
                }
            ],
        }

        if response_format:
            api_kwargs["response_format"] = response_format

        response = self._client.chat.completions.create(**api_kwargs)

        return response.choices[0].message.content  # type: ignore[return-value,no-any-return]

    def query_batch(
        self, images: list[Image], query: str, response_format: dict[str, Any] | None = None, **kwargs: Any
    ) -> list[str]:  # type: ignore[override]
        """Query VLM with multiple images using a single API call."""
        if not images:
            return []

        content: list[dict[str, Any]] = [
            {
                "type": "image_url",
                "image_url": {"url": f"data:image/png;base64,{self._prepare_image(img)[0].to_base64()}"},
            }
            for img in images
        ]
        content.append({"type": "text", "text": query})

        messages = [{"role": "user", "content": content}]
        api_kwargs: dict[str, Any] = {"model": self.config.model_name, "messages": messages}
        if response_format:
            api_kwargs["response_format"] = response_format

        response = self._client.chat.completions.create(**api_kwargs)
        response_text = response.choices[0].message.content or ""
        # Return one response per image (same response since API analyzes all images together)
        return [response_text] * len(images)

    def stop(self) -> None:
        """Release the OpenAI client."""
        if "_client" in self.__dict__:
            del self.__dict__["_client"]




# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from dimos.agents.agent import Agent
from dimos.protocol.skill.test_coordinator import SkillContainerTest


@pytest.mark.tool
@pytest.mark.asyncio
async def test_agent_init() -> None:
    system_prompt = (
        "Your name is Mr. Potato, potatoes are bad at math. Use a tools if asked to calculate"
    )

    # # Uncomment the following lines to use a dimos module system
    # dimos = start(2)
    # testcontainer = dimos.deploy(SkillContainerTest)
    # agent = Agent(system_prompt=system_prompt)

    ## uncomment the following lines to run agents in a main loop without a module system
    testcontainer = SkillContainerTest()
    agent = Agent(system_prompt=system_prompt)

    agent.register_skills(testcontainer)
    agent.start()

    agent.run_implicit_skill("uptime_seconds")

    await agent.query_async(
        "hi there, please tell me what's your name and current date, and how much is 124181112 + 124124?"
    )

    # agent loop is considered finished once no active skills remain,
    # agent will stop it's loop if passive streams are active
    print("Agent loop finished, asking about camera")

    # we query again (this shows subsequent querying, but we could have asked for camera image in the original query,
    # it all runs in parallel, and agent might get called once or twice depending on timing of skill responses)
    # await agent.query_async("tell me what you see on the camera?")

    # you can run skillspy and agentspy in parallel with this test for a better observation of what's happening
    await agent.query_async("tell me exactly everything we've talked about until now")

    print("Agent loop finished")

    agent.stop()
    testcontainer.stop()
    dimos.stop()



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import queue
import threading
import time
from typing import Any

from pydantic import Field
from reactivex import Subject

from dimos.skills.skills import AbstractSkill
from dimos.utils.logging_config import setup_logger

logger = setup_logger()

# Global lock to prevent multiple simultaneous audio playbacks
_audio_device_lock = threading.RLock()

# Global queue for sequential audio processing
_audio_queue = queue.Queue()  # type: ignore[var-annotated]
_queue_processor_thread = None
_queue_running = False


def _process_audio_queue() -> None:
    """Background thread to process audio requests sequentially"""
    global _queue_running

    while _queue_running:
        try:
            # Get the next queued audio task with a timeout
            task = _audio_queue.get(timeout=1.0)
            if task is None:  # Sentinel value to stop the thread
                break

            # Execute the task (which is a function to be called)
            task()
            _audio_queue.task_done()

        except queue.Empty:
            # No tasks in queue, just continue waiting
            continue
        except Exception as e:
            logger.error(f"Error in audio queue processor: {e}")
            # Continue processing other tasks


def start_audio_queue_processor() -> None:
    """Start the background thread for processing audio requests"""
    global _queue_processor_thread, _queue_running

    if _queue_processor_thread is None or not _queue_processor_thread.is_alive():
        _queue_running = True
        _queue_processor_thread = threading.Thread(
            target=_process_audio_queue, daemon=True, name="AudioQueueProcessor"
        )
        _queue_processor_thread.start()
        logger.info("Started audio queue processor thread")


# Start the queue processor when module is imported
start_audio_queue_processor()


class Speak(AbstractSkill):
    """Speak text out loud to humans nearby or to other robots."""

    text: str = Field(..., description="Text to speak")

    def __init__(self, tts_node: Any | None = None, **data) -> None:  # type: ignore[no-untyped-def]
        super().__init__(**data)
        self._tts_node = tts_node
        self._audio_complete = threading.Event()
        self._subscription = None
        self._subscriptions: list = []  # type: ignore[type-arg]  # Track all subscriptions

    def __call__(self):  # type: ignore[no-untyped-def]
        if not self._tts_node:
            logger.error("No TTS node provided to Speak skill")
            return "Error: No TTS node available"

        # Create a result queue to get the result back from the audio thread
        result_queue = queue.Queue(1)  # type: ignore[var-annotated]

        # Define the speech task to run in the audio queue
        def speak_task() -> None:
            try:
                # Using a lock to ensure exclusive access to audio device
                with _audio_device_lock:
                    text_subject = Subject()  # type: ignore[var-annotated]
                    self._audio_complete.clear()
                    self._subscriptions = []

                    # This function will be called when audio processing is complete
                    def on_complete() -> None:
                        logger.info(f"TTS audio playback completed for: {self.text}")
                        self._audio_complete.set()

                    # This function will be called if there's an error
                    def on_error(error) -> None:  # type: ignore[no-untyped-def]
                        logger.error(f"Error in TTS processing: {error}")
                        self._audio_complete.set()

                    # Connect the Subject to the TTS node and keep the subscription
                    self._tts_node.consume_text(text_subject)  # type: ignore[union-attr]

                    # Subscribe to the audio output to know when it's done
                    self._subscription = self._tts_node.emit_text().subscribe(  # type: ignore[union-attr]
                        on_next=lambda text: logger.debug(f"TTS processing: {text}"),
                        on_completed=on_complete,
                        on_error=on_error,
                    )
                    self._subscriptions.append(self._subscription)

                    # Emit the text to the Subject
                    text_subject.on_next(self.text)
                    text_subject.on_completed()  # Signal that we're done sending text

                    # Wait for audio playback to complete with a timeout
                    # Using a dynamic timeout based on text length
                    timeout = max(5, len(self.text) * 0.1)
                    logger.debug(f"Waiting for TTS completion with timeout {timeout:.1f}s")

                    if not self._audio_complete.wait(timeout=timeout):
                        logger.warning(f"TTS timeout reached for: {self.text}")
                    else:
                        # Add a small delay after audio completes to ensure buffers are fully flushed
                        time.sleep(0.3)

                    # Clean up all subscriptions
                    for sub in self._subscriptions:
                        if sub:
                            sub.dispose()
                    self._subscriptions = []

                    # Successfully completed
                    result_queue.put(f"Spoke: {self.text} successfully")
            except Exception as e:
                logger.error(f"Error in speak task: {e}")
                result_queue.put(f"Error speaking text: {e!s}")

        # Add our speech task to the global queue for sequential processing
        display_text = self.text[:50] + "..." if len(self.text) > 50 else self.text
        logger.info(f"Queueing speech task: '{display_text}'")
        _audio_queue.put(speak_task)

        # Wait for the result with a timeout
        try:
            # Use a longer timeout than the audio playback itself
            text_len_timeout = len(self.text) * 0.15  # 150ms per character
            max_timeout = max(10, text_len_timeout)  # At least 10 seconds

            return result_queue.get(timeout=max_timeout)
        except queue.Empty:
            logger.error("Timed out waiting for speech task to complete")
            return f"Error: Timed out while speaking: {self.text}"



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ControlOrchestrator module.

Centralized control orchestrator that replaces per-driver/per-controller
loops with a single deterministic tick-based system.

Features:
- Single tick loop (read → compute → arbitrate → route → write)
- Per-joint arbitration (highest priority wins)
- Mode conflict detection
- Partial command support (hold last value)
- Aggregated preemption notifications
"""

from __future__ import annotations

from dataclasses import dataclass, field
import threading
import time
from typing import TYPE_CHECKING, Any

from dimos.control.hardware_interface import BackendHardwareInterface, HardwareInterface
from dimos.control.task import ControlTask
from dimos.control.tick_loop import TickLoop
from dimos.core import Module, Out, rpc
from dimos.core.module import ModuleConfig
from dimos.msgs.sensor_msgs import (
    JointState,  # noqa: TC001 - needed at runtime for Out[JointState]
)
from dimos.msgs.trajectory_msgs import JointTrajectory, TrajectoryState
from dimos.utils.logging_config import setup_logger

if TYPE_CHECKING:
    from dimos.hardware.manipulators.spec import ManipulatorBackend

logger = setup_logger()


# =============================================================================
# Configuration
# =============================================================================


@dataclass
class HardwareConfig:
    """Configuration for a hardware backend.

    Attributes:
        id: Unique hardware identifier (e.g., "arm", "left_arm")
        type: Backend type ("mock", "xarm", "piper")
        dof: Degrees of freedom
        joint_prefix: Prefix for joint names (defaults to id)
        ip: IP address (required for xarm)
        can_port: CAN port (for piper, default "can0")
        auto_enable: Whether to auto-enable servos (default True)
    """

    id: str
    type: str = "mock"
    dof: int = 7
    joint_prefix: str | None = None
    ip: str | None = None
    can_port: str | None = None
    auto_enable: bool = True


@dataclass
class TaskConfig:
    """Configuration for a control task.

    Attributes:
        name: Task name (e.g., "traj_arm")
        type: Task type ("trajectory")
        joint_names: List of joint names this task controls
        priority: Task priority (higher wins arbitration)
    """

    name: str
    type: str = "trajectory"
    joint_names: list[str] = field(default_factory=lambda: [])
    priority: int = 10


@dataclass
class TaskStatus:
    """Status of a control task.

    Attributes:
        active: Whether the task is currently active
        state: Task state name (e.g., "IDLE", "RUNNING", "DONE")
        progress: Task progress from 0.0 to 1.0
    """

    active: bool
    state: str | None = None
    progress: float | None = None


@dataclass
class ControlOrchestratorConfig(ModuleConfig):
    """Configuration for the ControlOrchestrator.

    Attributes:
        tick_rate: Control loop frequency in Hz (default: 100)
        publish_joint_state: Whether to publish aggregated JointState
        joint_state_frame_id: Frame ID for published JointState
        log_ticks: Whether to log tick information (verbose)
        hardware: List of hardware configurations to create on start
        tasks: List of task configurations to create on start
    """

    tick_rate: float = 100.0
    publish_joint_state: bool = True
    joint_state_frame_id: str = "orchestrator"
    log_ticks: bool = False
    hardware: list[HardwareConfig] = field(default_factory=lambda: [])
    tasks: list[TaskConfig] = field(default_factory=lambda: [])


# =============================================================================
# ControlOrchestrator Module
# =============================================================================


class ControlOrchestrator(Module[ControlOrchestratorConfig]):
    """Centralized control orchestrator with per-joint arbitration.

    Single tick loop that:
    1. Reads state from all hardware
    2. Runs all active tasks
    3. Arbitrates conflicts per-joint (highest priority wins)
    4. Routes commands to hardware
    5. Publishes aggregated joint state

    Key design decisions:
    - Joint-centric commands (not hardware-centric)
    - Per-joint arbitration (not per-hardware)
    - Centralized time (tasks use state.t_now, never time.time())
    - Partial commands OK (hardware holds last value)
    - Aggregated preemption (one notification per task per tick)

    Example:
        >>> from dimos.control import ControlOrchestrator
        >>> from dimos.hardware.manipulators.xarm import XArmBackend
        >>>
        >>> orch = ControlOrchestrator(tick_rate=100.0)
        >>> backend = XArmBackend(ip="192.168.1.185", dof=7)
        >>> backend.connect()
        >>> orch.add_hardware("left_arm", backend, joint_prefix="left")
        >>> orch.start()
    """

    # Output: Aggregated joint state for external consumers
    joint_state: Out[JointState]

    config: ControlOrchestratorConfig
    default_config = ControlOrchestratorConfig

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        # Hardware interfaces (keyed by hardware_id)
        self._hardware: dict[str, HardwareInterface] = {}
        self._hardware_lock = threading.Lock()

        # Joint -> hardware mapping (built when hardware added)
        self._joint_to_hardware: dict[str, str] = {}

        # Registered tasks
        self._tasks: dict[str, ControlTask] = {}
        self._task_lock = threading.Lock()

        # Tick loop (created on start)
        self._tick_loop: TickLoop | None = None

        logger.info(f"ControlOrchestrator initialized at {self.config.tick_rate}Hz")

    # =========================================================================
    # Config-based Setup
    # =========================================================================

    def _setup_from_config(self) -> None:
        """Create hardware and tasks from config (called on start)."""
        hardware_added: list[str] = []

        try:
            for hw_cfg in self.config.hardware:
                self._setup_hardware(hw_cfg)
                hardware_added.append(hw_cfg.id)

            for task_cfg in self.config.tasks:
                task = self._create_task_from_config(task_cfg)
                self.add_task(task)

        except Exception:
            # Rollback: clean up all successfully added hardware
            for hw_id in hardware_added:
                try:
                    self.remove_hardware(hw_id)
                except Exception:
                    pass
            raise

    def _setup_hardware(self, hw_cfg: HardwareConfig) -> None:
        """Connect and add a single hardware backend."""
        backend = self._create_backend_from_config(hw_cfg)

        if not backend.connect():
            raise RuntimeError(f"Failed to connect to {hw_cfg.type} backend")

        try:
            if hw_cfg.auto_enable and hasattr(backend, "write_enable"):
                backend.write_enable(True)
            self.add_hardware(
                hw_cfg.id,
                backend,
                joint_prefix=hw_cfg.joint_prefix or hw_cfg.id,
            )
        except Exception:
            backend.disconnect()
            raise

    def _create_backend_from_config(self, cfg: HardwareConfig) -> ManipulatorBackend:
        """Create a manipulator backend from config."""
        match cfg.type.lower():
            case "mock":
                from dimos.hardware.manipulators.mock import MockBackend

                return MockBackend(dof=cfg.dof)
            case "xarm":
                if cfg.ip is None:
                    raise ValueError("ip is required for xarm backend")
                from dimos.hardware.manipulators.xarm import XArmBackend

                return XArmBackend(ip=cfg.ip, dof=cfg.dof)
            case "piper":
                from dimos.hardware.manipulators.piper import PiperBackend

                return PiperBackend(can_port=cfg.can_port or "can0", dof=cfg.dof)
            case _:
                raise ValueError(f"Unknown backend type: {cfg.type}")

    def _create_task_from_config(self, cfg: TaskConfig) -> ControlTask:
        """Create a control task from config."""
        task_type = cfg.type.lower()

        if task_type == "trajectory":
            from dimos.control.tasks import JointTrajectoryTask, JointTrajectoryTaskConfig

            return JointTrajectoryTask(
                cfg.name,
                JointTrajectoryTaskConfig(
                    joint_names=cfg.joint_names,
                    priority=cfg.priority,
                ),
            )

        else:
            raise ValueError(f"Unknown task type: {task_type}")

    # =========================================================================
    # Hardware Management (RPC)
    # =========================================================================

    @rpc
    def add_hardware(
        self,
        hardware_id: str,
        backend: ManipulatorBackend,
        joint_prefix: str | None = None,
    ) -> bool:
        """Register a hardware backend with the orchestrator."""
        with self._hardware_lock:
            if hardware_id in self._hardware:
                logger.warning(f"Hardware {hardware_id} already registered")
                return False

            interface = BackendHardwareInterface(
                backend=backend,
                hardware_id=hardware_id,
                joint_prefix=joint_prefix,
            )
            self._hardware[hardware_id] = interface

            for joint_name in interface.joint_names:
                self._joint_to_hardware[joint_name] = hardware_id

            logger.info(f"Added hardware {hardware_id} with joints: {interface.joint_names}")
            return True

    @rpc
    def remove_hardware(self, hardware_id: str) -> bool:
        """Remove a hardware interface.

        Note: For safety, call this only when no tasks are actively using this
        hardware. Consider stopping the orchestrator before removing hardware.
        """
        with self._hardware_lock:
            if hardware_id not in self._hardware:
                return False

            interface = self._hardware[hardware_id]
            hw_joints = set(interface.joint_names)

            with self._task_lock:
                for task in self._tasks.values():
                    if task.is_active():
                        claimed_joints = task.claim().joints
                        overlap = hw_joints & claimed_joints
                        if overlap:
                            logger.error(
                                f"Cannot remove hardware {hardware_id}: "
                                f"task '{task.name}' is actively using joints {overlap}"
                            )
                            return False

            for joint_name in interface.joint_names:
                del self._joint_to_hardware[joint_name]

            interface.disconnect()
            del self._hardware[hardware_id]
            logger.info(f"Removed hardware {hardware_id}")
            return True

    @rpc
    def list_hardware(self) -> list[str]:
        """List registered hardware IDs."""
        with self._hardware_lock:
            return list(self._hardware.keys())

    @rpc
    def list_joints(self) -> list[str]:
        """List all joint names across all hardware."""
        with self._hardware_lock:
            return list(self._joint_to_hardware.keys())

    @rpc
    def get_joint_positions(self) -> dict[str, float]:
        """Get current joint positions for all joints."""
        with self._hardware_lock:
            positions: dict[str, float] = {}
            for hw in self._hardware.values():
                state = hw.read_state()  # {joint_name: (pos, vel, effort)}
                for joint_name, (pos, _vel, _effort) in state.items():
                    positions[joint_name] = pos
            return positions

    # =========================================================================
    # Task Management (RPC)
    # =========================================================================

    @rpc
    def add_task(self, task: ControlTask) -> bool:
        """Register a task with the orchestrator."""
        if not isinstance(task, ControlTask):
            raise TypeError("task must implement ControlTask")

        with self._task_lock:
            if task.name in self._tasks:
                logger.warning(f"Task {task.name} already registered")
                return False
            self._tasks[task.name] = task
            logger.info(f"Added task {task.name}")
            return True

    @rpc
    def remove_task(self, task_name: str) -> bool:
        """Remove a task by name."""
        with self._task_lock:
            if task_name in self._tasks:
                del self._tasks[task_name]
                logger.info(f"Removed task {task_name}")
                return True
            return False

    @rpc
    def get_task(self, task_name: str) -> ControlTask | None:
        """Get a task by name."""
        with self._task_lock:
            return self._tasks.get(task_name)

    @rpc
    def list_tasks(self) -> list[str]:
        """List registered task names."""
        with self._task_lock:
            return list(self._tasks.keys())

    @rpc
    def get_active_tasks(self) -> list[str]:
        """List currently active task names."""
        with self._task_lock:
            return [name for name, task in self._tasks.items() if task.is_active()]

    # =========================================================================
    # Trajectory Execution (RPC)
    # =========================================================================

    @rpc
    def execute_trajectory(self, task_name: str, trajectory: JointTrajectory) -> bool:
        """Execute a trajectory on a named task."""
        with self._task_lock:
            task = self._tasks.get(task_name)
            if task is None:
                logger.warning(f"Task {task_name} not found")
                return False

            if not hasattr(task, "execute"):
                logger.warning(f"Task {task_name} doesn't support execute()")
                return False

            logger.info(
                f"Executing trajectory on {task_name}: "
                f"{len(trajectory.points)} points, duration={trajectory.duration:.3f}s"
            )
            return task.execute(trajectory)  # type: ignore[attr-defined,no-any-return]

    @rpc
    def get_trajectory_status(self, task_name: str) -> TaskStatus | None:
        """Get the status of a trajectory task."""
        with self._task_lock:
            task = self._tasks.get(task_name)
            if task is None:
                return None

            state: str | None = None
            if hasattr(task, "get_state"):
                task_state: TrajectoryState = task.get_state()  # type: ignore[attr-defined]
                state = (
                    task_state.name if isinstance(task_state, TrajectoryState) else str(task_state)
                )

            progress: float | None = None
            if hasattr(task, "get_progress"):
                t_now = time.perf_counter()
                progress = task.get_progress(t_now)  # type: ignore[attr-defined]

            return TaskStatus(active=task.is_active(), state=state, progress=progress)

    @rpc
    def cancel_trajectory(self, task_name: str) -> bool:
        """Cancel an active trajectory on a task."""
        with self._task_lock:
            task = self._tasks.get(task_name)
            if task is None:
                logger.warning(f"Task {task_name} not found")
                return False

            if not hasattr(task, "cancel"):
                logger.warning(f"Task {task_name} doesn't support cancel()")
                return False

            logger.info(f"Cancelling trajectory on {task_name}")
            return task.cancel()  # type: ignore[attr-defined,no-any-return]

    # =========================================================================
    # Lifecycle
    # =========================================================================

    @rpc
    def start(self) -> None:
        """Start the orchestrator control loop."""
        if self._tick_loop and self._tick_loop.is_running:
            logger.warning("Orchestrator already running")
            return

        super().start()

        # Setup hardware and tasks from config (if any)
        if self.config.hardware or self.config.tasks:
            self._setup_from_config()

        # Create and start tick loop
        publish_cb = self.joint_state.publish if self.config.publish_joint_state else None
        self._tick_loop = TickLoop(
            tick_rate=self.config.tick_rate,
            hardware=self._hardware,
            hardware_lock=self._hardware_lock,
            tasks=self._tasks,
            task_lock=self._task_lock,
            joint_to_hardware=self._joint_to_hardware,
            publish_callback=publish_cb,
            frame_id=self.config.joint_state_frame_id,
            log_ticks=self.config.log_ticks,
        )
        self._tick_loop.start()

        logger.info(f"ControlOrchestrator started at {self.config.tick_rate}Hz")

    @rpc
    def stop(self) -> None:
        """Stop the orchestrator."""
        logger.info("Stopping ControlOrchestrator...")

        if self._tick_loop:
            self._tick_loop.stop()

        # Disconnect all hardware backends
        with self._hardware_lock:
            for hw_id, interface in self._hardware.items():
                try:
                    interface.disconnect()
                    logger.info(f"Disconnected hardware {hw_id}")
                except Exception as e:
                    logger.error(f"Error disconnecting hardware {hw_id}: {e}")

        super().stop()
        logger.info("ControlOrchestrator stopped")

    @rpc
    def get_tick_count(self) -> int:
        """Get the number of ticks since start."""
        return self._tick_loop.tick_count if self._tick_loop else 0


# Blueprint export
control_orchestrator = ControlOrchestrator.blueprint


__all__ = [
    "ControlOrchestrator",
    "ControlOrchestratorConfig",
    "HardwareConfig",
    "TaskConfig",
    "control_orchestrator",
]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dimos_lcm.geometry_msgs import Twist as LCMTwist
from plum import dispatch

try:
    from geometry_msgs.msg import (  # type: ignore[attr-defined]
        Twist as ROSTwist,
        Vector3 as ROSVector3,
    )
except ImportError:
    ROSTwist = None  # type: ignore[assignment, misc]
    ROSVector3 = None  # type: ignore[assignment, misc]

# Import Quaternion at runtime for beartype compatibility
# (beartype needs to resolve forward references at runtime)
from dimos.msgs.geometry_msgs.Quaternion import Quaternion
from dimos.msgs.geometry_msgs.Vector3 import Vector3, VectorLike


class Twist(LCMTwist):  # type: ignore[misc]
    linear: Vector3
    angular: Vector3
    msg_name = "geometry_msgs.Twist"

    @dispatch
    def __init__(self) -> None:
        """Initialize a zero twist (no linear or angular velocity)."""
        self.linear = Vector3()
        self.angular = Vector3()

    @dispatch  # type: ignore[no-redef]
    def __init__(self, linear: VectorLike, angular: VectorLike) -> None:
        """Initialize a twist from linear and angular velocities."""

        self.linear = Vector3(linear)
        self.angular = Vector3(angular)

    @dispatch  # type: ignore[no-redef]
    def __init__(self, linear: VectorLike, angular: Quaternion) -> None:
        """Initialize a twist from linear velocity and angular as quaternion (converted to euler)."""
        self.linear = Vector3(linear)
        self.angular = angular.to_euler()

    @dispatch  # type: ignore[no-redef]
    def __init__(self, twist: Twist) -> None:
        """Initialize from another Twist (copy constructor)."""
        self.linear = Vector3(twist.linear)
        self.angular = Vector3(twist.angular)

    @dispatch  # type: ignore[no-redef]
    def __init__(self, lcm_twist: LCMTwist) -> None:
        """Initialize from an LCM Twist."""
        self.linear = Vector3(lcm_twist.linear)
        self.angular = Vector3(lcm_twist.angular)

    @dispatch  # type: ignore[no-redef]
    def __init__(self, **kwargs) -> None:
        """Handle keyword arguments for LCM compatibility."""
        linear = kwargs.get("linear", Vector3())
        angular = kwargs.get("angular", Vector3())

        self.__init__(linear, angular)

    def __repr__(self) -> str:
        return f"Twist(linear={self.linear!r}, angular={self.angular!r})"

    def __str__(self) -> str:
        return f"Twist:\n  Linear: {self.linear}\n  Angular: {self.angular}"

    def __eq__(self, other) -> bool:  # type: ignore[no-untyped-def]
        """Check if two twists are equal."""
        if not isinstance(other, Twist):
            return False
        return self.linear == other.linear and self.angular == other.angular

    @classmethod
    def zero(cls) -> Twist:
        """Create a zero twist (no motion)."""
        return cls()

    def is_zero(self) -> bool:
        """Check if this is a zero twist (no linear or angular velocity)."""
        return self.linear.is_zero() and self.angular.is_zero()

    def __bool__(self) -> bool:
        """Boolean conversion for Twist.

        A Twist is considered False if it's a zero twist (no motion),
        and True otherwise.

        Returns:
            False if twist is zero, True otherwise
        """
        return not self.is_zero()

    @classmethod
    def from_ros_msg(cls, ros_msg: ROSTwist) -> Twist:
        """Create a Twist from a ROS geometry_msgs/Twist message.

        Args:
            ros_msg: ROS Twist message

        Returns:
            Twist instance
        """

        linear = Vector3(ros_msg.linear.x, ros_msg.linear.y, ros_msg.linear.z)
        angular = Vector3(ros_msg.angular.x, ros_msg.angular.y, ros_msg.angular.z)
        return cls(linear, angular)

    def to_ros_msg(self) -> ROSTwist:
        """Convert to a ROS geometry_msgs/Twist message.

        Returns:
            ROS Twist message
        """

        ros_msg = ROSTwist()  # type: ignore[no-untyped-call]
        ros_msg.linear = ROSVector3(x=self.linear.x, y=self.linear.y, z=self.linear.z)  # type: ignore[no-untyped-call]
        ros_msg.angular = ROSVector3(x=self.angular.x, y=self.angular.y, z=self.angular.z)  # type: ignore[no-untyped-call]
        return ros_msg


__all__ = ["Quaternion", "Twist"]



#!/usr/bin/env python3

# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import atexit
import base64
from collections.abc import Callable
import functools
import json
import pickle
import subprocess
import sys
import threading
import time
from typing import Any, TypeVar
import weakref

import numpy as np
from numpy.typing import NDArray
from reactivex import Observable
from reactivex.abc import ObserverBase, SchedulerBase
from reactivex.disposable import Disposable

from dimos.core.global_config import GlobalConfig
from dimos.msgs.geometry_msgs import Quaternion, Twist, Vector3
from dimos.msgs.sensor_msgs import CameraInfo, Image, ImageFormat, PointCloud2
from dimos.robot.unitree_webrtc.type.odometry import Odometry
from dimos.simulation.mujoco.constants import (
    LAUNCHER_PATH,
    LIDAR_FPS,
    VIDEO_CAMERA_FOV,
    VIDEO_FPS,
    VIDEO_HEIGHT,
    VIDEO_WIDTH,
)
from dimos.simulation.mujoco.shared_memory import ShmWriter
from dimos.utils.data import get_data
from dimos.utils.logging_config import setup_logger

ODOM_FREQUENCY = 50

logger = setup_logger()

T = TypeVar("T")


class MujocoConnection:
    """MuJoCo simulator connection that runs in a separate subprocess."""

    def __init__(self, global_config: GlobalConfig) -> None:
        try:
            import mujoco
        except ImportError:
            raise ImportError("'mujoco' is not installed. Use `pip install -e .[sim]`")

        # Pre-download the mujoco_sim data.
        get_data("mujoco_sim")

        # Trigger the download of the mujoco_menagerie package. This is so it
        # doesn't trigger in the mujoco process where it can time out.
        from mujoco_playground._src import mjx_env

        mjx_env.ensure_menagerie_exists()

        self.global_config = global_config
        self.process: subprocess.Popen[bytes] | None = None
        self.shm_data: ShmWriter | None = None
        self._last_video_seq = 0
        self._last_odom_seq = 0
        self._last_lidar_seq = 0
        self._stop_timer: threading.Timer | None = None

        self._stream_threads: list[threading.Thread] = []
        self._stop_events: list[threading.Event] = []
        self._is_cleaned_up = False

    @staticmethod
    def _compute_camera_info() -> CameraInfo:
        """Compute camera intrinsics from MuJoCo camera parameters.

        Uses pinhole camera model: f = height / (2 * tan(fovy / 2))
        """
        import math

        fovy = math.radians(VIDEO_CAMERA_FOV)
        f = VIDEO_HEIGHT / (2 * math.tan(fovy / 2))
        cx = VIDEO_WIDTH / 2.0
        cy = VIDEO_HEIGHT / 2.0

        return CameraInfo(
            frame_id="camera_optical",
            height=VIDEO_HEIGHT,
            width=VIDEO_WIDTH,
            distortion_model="plumb_bob",
            D=[0.0, 0.0, 0.0, 0.0, 0.0],
            K=[f, 0.0, cx, 0.0, f, cy, 0.0, 0.0, 1.0],
            R=[1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
            P=[f, 0.0, cx, 0.0, 0.0, f, cy, 0.0, 0.0, 0.0, 1.0, 0.0],
        )

    camera_info_static: CameraInfo = _compute_camera_info()

    def start(self) -> None:
        self.shm_data = ShmWriter()

        config_pickle = base64.b64encode(pickle.dumps(self.global_config)).decode("ascii")
        shm_names_json = json.dumps(self.shm_data.shm.to_names())

        # Launch the subprocess
        try:
            # mjpython must be used macOS (because of launch_passive inside mujoco_process.py)
            executable = sys.executable if sys.platform != "darwin" else "mjpython"

            self.process = subprocess.Popen(
                [executable, str(LAUNCHER_PATH), config_pickle, shm_names_json],
            )

        except Exception as e:
            self.shm_data.cleanup()
            raise RuntimeError(f"Failed to start MuJoCo subprocess: {e}") from e

        # Wait for process to be ready
        ready_timeout = 300.0
        start_time = time.time()
        assert self.process is not None
        while time.time() - start_time < ready_timeout:
            if self.process.poll() is not None:
                exit_code = self.process.returncode
                self.stop()
                raise RuntimeError(f"MuJoCo process failed to start (exit code {exit_code})")
            if self.shm_data.is_ready():
                logger.info("MuJoCo process started successfully")
                # Register atexit handler to ensure subprocess is cleaned up
                # Use weakref to avoid preventing garbage collection
                weak_self = weakref.ref(self)

                def cleanup_on_exit() -> None:
                    instance = weak_self()
                    if instance is not None:
                        instance.stop()

                atexit.register(cleanup_on_exit)
                return
            time.sleep(0.1)

        # Timeout
        self.stop()
        raise RuntimeError("MuJoCo process failed to start (timeout)")

    def stop(self) -> None:
        if self._is_cleaned_up:
            return

        self._is_cleaned_up = True

        # clean up open file descriptors
        if self.process:
            if self.process.stderr:
                self.process.stderr.close()
            if self.process.stdout:
                self.process.stdout.close()

        # Cancel any pending timers
        if self._stop_timer:
            self._stop_timer.cancel()
            self._stop_timer = None

        # Stop all stream threads
        for stop_event in self._stop_events:
            stop_event.set()

        # Wait for threads to finish
        for thread in self._stream_threads:
            if thread.is_alive():
                thread.join(timeout=2.0)
                if thread.is_alive():
                    logger.warning(f"Stream thread {thread.name} did not stop gracefully")

        # Signal subprocess to stop
        if self.shm_data:
            self.shm_data.signal_stop()

        # Wait for process to finish
        if self.process:
            try:
                self.process.terminate()
                try:
                    self.process.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    logger.warning("MuJoCo process did not stop gracefully, killing")
                    self.process.kill()
                    self.process.wait(timeout=2)
            except Exception as e:
                logger.error(f"Error stopping MuJoCo process: {e}")

            self.process = None

        # Clean up shared memory
        if self.shm_data:
            self.shm_data.cleanup()
            self.shm_data = None

        # Clear references
        self._stream_threads.clear()
        self._stop_events.clear()

        self.lidar_stream.cache_clear()
        self.odom_stream.cache_clear()
        self.video_stream.cache_clear()

    def standup(self) -> bool:
        return True

    def liedown(self) -> bool:
        return True

    def get_video_frame(self) -> NDArray[Any] | None:
        if self.shm_data is None:
            return None

        frame, seq = self.shm_data.read_video()
        if seq > self._last_video_seq:
            self._last_video_seq = seq
            return frame

        return None

    def get_odom_message(self) -> Odometry | None:
        if self.shm_data is None:
            return None

        odom_data, seq = self.shm_data.read_odom()
        if seq > self._last_odom_seq and odom_data is not None:
            self._last_odom_seq = seq
            pos, quat_wxyz, timestamp = odom_data

            # Convert quaternion from (w,x,y,z) to (x,y,z,w) for ROS/Dimos
            orientation = Quaternion(quat_wxyz[1], quat_wxyz[2], quat_wxyz[3], quat_wxyz[0])

            return Odometry(
                position=Vector3(pos[0], pos[1], pos[2]),
                orientation=orientation,
                ts=timestamp,
                frame_id="world",
            )

        return None

    def get_lidar_message(self) -> PointCloud2 | None:
        if self.shm_data is None:
            return None

        lidar_msg, seq = self.shm_data.read_lidar()
        if seq > self._last_lidar_seq and lidar_msg is not None:
            self._last_lidar_seq = seq
            return lidar_msg

        return None

    def _create_stream(
        self,
        getter: Callable[[], T | None],
        frequency: float,
        stream_name: str,
    ) -> Observable[T]:
        def on_subscribe(observer: ObserverBase[T], _scheduler: SchedulerBase | None) -> Disposable:
            if self._is_cleaned_up:
                observer.on_completed()
                return Disposable(lambda: None)

            stop_event = threading.Event()
            self._stop_events.append(stop_event)

            def run() -> None:
                try:
                    while not stop_event.is_set() and not self._is_cleaned_up:
                        data = getter()
                        if data is not None:
                            observer.on_next(data)
                        time.sleep(1 / frequency)
                except Exception as e:
                    logger.error(f"{stream_name} stream error: {e}")
                finally:
                    observer.on_completed()

            thread = threading.Thread(target=run, daemon=True)
            self._stream_threads.append(thread)
            thread.start()

            def dispose() -> None:
                stop_event.set()

            return Disposable(dispose)

        return Observable(on_subscribe)

    @functools.cache
    def lidar_stream(self) -> Observable[PointCloud2]:
        return self._create_stream(self.get_lidar_message, LIDAR_FPS, "Lidar")

    @functools.cache
    def odom_stream(self) -> Observable[Odometry]:
        return self._create_stream(self.get_odom_message, ODOM_FREQUENCY, "Odom")

    @functools.cache
    def video_stream(self) -> Observable[Image]:
        def get_video_as_image() -> Image | None:
            frame = self.get_video_frame()
            # MuJoCo renderer returns RGB uint8 frames; Image.from_numpy defaults to BGR.
            return Image.from_numpy(frame, format=ImageFormat.RGB) if frame is not None else None

        return self._create_stream(get_video_as_image, VIDEO_FPS, "Video")

    def move(self, twist: Twist, duration: float = 0.0) -> bool:
        if self._is_cleaned_up or self.shm_data is None:
            return True

        linear = np.array([twist.linear.x, twist.linear.y, twist.linear.z], dtype=np.float32)
        angular = np.array([twist.angular.x, twist.angular.y, twist.angular.z], dtype=np.float32)
        self.shm_data.write_command(linear, angular)

        if duration > 0:
            if self._stop_timer:
                self._stop_timer.cancel()

            def stop_movement() -> None:
                if self.shm_data:
                    self.shm_data.write_command(
                        np.zeros(3, dtype=np.float32), np.zeros(3, dtype=np.float32)
                    )
                self._stop_timer = None

            self._stop_timer = threading.Timer(duration, stop_movement)
            self._stop_timer.daemon = True
            self._stop_timer.start()
        return True

    def publish_request(self, topic: str, data: dict[str, Any]) -> dict[Any, Any]:
        print(f"publishing request, topic={topic}, data={data}")
        return {}



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
import time
from typing import Any

import open3d as o3d  # type: ignore[import-untyped]
from reactivex import interval
from reactivex.disposable import Disposable

from dimos.core import DimosCluster, In, LCMTransport, Module, Out, rpc
from dimos.core.global_config import GlobalConfig
from dimos.mapping.pointclouds.accumulators.general import GeneralPointCloudAccumulator
from dimos.mapping.pointclouds.accumulators.protocol import PointCloudAccumulator
from dimos.mapping.pointclouds.occupancy import general_occupancy
from dimos.msgs.nav_msgs import OccupancyGrid
from dimos.msgs.sensor_msgs import PointCloud2
from dimos.robot.unitree.connection.go2 import Go2ConnectionProtocol


class Map(Module):
    lidar: In[PointCloud2]
    global_map: Out[PointCloud2]
    global_costmap: Out[OccupancyGrid]

    _point_cloud_accumulator: PointCloudAccumulator
    _global_config: GlobalConfig
    _preloaded_occupancy: OccupancyGrid | None = None

    def __init__(  # type: ignore[no-untyped-def]
        self,
        voxel_size: float = 0.05,
        cost_resolution: float = 0.05,
        global_publish_interval: float | None = None,
        min_height: float = 0.10,
        max_height: float = 0.5,
        global_config: GlobalConfig | None = None,
        **kwargs,
    ) -> None:
        self.voxel_size = voxel_size
        self.cost_resolution = cost_resolution
        self.global_publish_interval = global_publish_interval
        self.min_height = min_height
        self.max_height = max_height
        self._global_config = global_config or GlobalConfig()
        self._point_cloud_accumulator = GeneralPointCloudAccumulator(
            self.voxel_size, self._global_config
        )

        if self._global_config.simulation:
            self.min_height = 0.3

        super().__init__(**kwargs)

    @rpc
    def start(self) -> None:
        super().start()

        self._disposables.add(Disposable(self.lidar.subscribe(self.add_frame)))

        if self.global_publish_interval is not None:
            unsub = interval(self.global_publish_interval).subscribe(self._publish)
            self._disposables.add(unsub)

    @rpc
    def stop(self) -> None:
        super().stop()

    def to_PointCloud2(self) -> PointCloud2:
        return PointCloud2(
            pointcloud=self._point_cloud_accumulator.get_point_cloud(),
            ts=time.time(),
        )

    # TODO: Why is this RPC?
    @rpc
    def add_frame(self, frame: PointCloud2) -> None:
        self._point_cloud_accumulator.add(frame.pointcloud)

    @property
    def o3d_geometry(self) -> o3d.geometry.PointCloud:
        return self._point_cloud_accumulator.get_point_cloud()

    def _publish(self, _: Any) -> None:
        self.global_map.publish(self.to_PointCloud2())

        occupancygrid = general_occupancy(
            self.to_PointCloud2(),
            resolution=self.cost_resolution,
            min_height=self.min_height,
            max_height=self.max_height,
        )

        # When debugging occupancy navigation, load a predefined occupancy grid.
        if self._global_config.mujoco_global_costmap_from_occupancy:
            if self._preloaded_occupancy is None:
                path = Path(self._global_config.mujoco_global_costmap_from_occupancy)
                self._preloaded_occupancy = OccupancyGrid.from_path(path)
            occupancygrid = self._preloaded_occupancy

        self.global_costmap.publish(occupancygrid)


mapper = Map.blueprint


def deploy(dimos: DimosCluster, connection: Go2ConnectionProtocol):  # type: ignore[no-untyped-def]
    mapper = dimos.deploy(Map, global_publish_interval=1.0)  # type: ignore[attr-defined]
    mapper.global_map.transport = LCMTransport("/global_map", PointCloud2)
    mapper.global_costmap.transport = LCMTransport("/global_costmap", OccupancyGrid)
    mapper.lidar.connect(connection.pointcloud)  # type: ignore[attr-defined]
    mapper.start()
    return mapper


__all__ = ["Map", "mapper"]
